@@ -1,78 +1,89 @@
+/*
+ * Copyright (C) 2013 Glyptodon LLC
+ *
+ * Permission is hereby granted, free of charge, to any person obtaining a copy
+ * of this software and associated documentation files (the "Software"), to deal
+ * in the Software without restriction, including without limitation the rights
+ * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+ * copies of the Software, and to permit persons to whom the Software is
+ * furnished to do so, subject to the following conditions:
+ *
+ * The above copyright notice and this permission notice shall be included in
+ * all copies or substantial portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+ * THE SOFTWARE.
+ */
 
-/* ***** BEGIN LICENSE BLOCK *****
- * Version: MPL 1.1/GPL 2.0/LGPL 2.1
- *
- * The contents of this file are subject to the Mozilla Public License Version
- * 1.1 (the "License"); you may not use this file except in compliance with
- * the License. You may obtain a copy of the License at
- * http://www.mozilla.org/MPL/
- *
- * Software distributed under the License is distributed on an "AS IS" basis,
- * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
- * for the specific language governing rights and limitations under the
- * License.
- *
- * The Original Code is guacd.
- *
- * The Initial Developer of the Original Code is
- * Michael Jumper.
- * Portions created by the Initial Developer are Copyright (C) 2010
- * the Initial Developer. All Rights Reserved.
- *
- * Contributor(s):
- * David PHAM-VAN <d.pham-van@ulteo.com> Ulteo SAS - http://www.ulteo.com
- * Alex Bligh <alex@alex.org.uk>
- *
- * Alternatively, the contents of this file may be used under the terms of
- * either the GNU General Public License Version 2 or later (the "GPL"), or
- * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
- * in which case the provisions of the GPL or the LGPL are applicable instead
- * of those above. If you wish to allow use of your version of this file only
- * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the MPL, indicate your
- * decision by deleting the provisions above and replace them with the notice
- * and other provisions required by the GPL or the LGPL. If you do not delete
- * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the MPL, the GPL or the LGPL.
- *
- * ***** END LICENSE BLOCK ***** */
+#include "config.h"
 
-#include <unistd.h>
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
-#include <signal.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <fcntl.h>
-#include <ctype.h>
-
-#include <sys/socket.h>
-#include <netdb.h>
-#include <netinet/in.h>
-
-#include <errno.h>
-#include <syslog.h>
-#include <libgen.h>
-
-#ifdef ENABLE_SSL
-#include <openssl/ssl.h>
-#include "socket-ssl.h"
-#endif
+#include "client.h"
+#include "client-map.h"
+#include "conf-args.h"
+#include "conf-file.h"
+#include "log.h"
 
 #include <guacamole/client.h>
 #include <guacamole/error.h>
 #include <guacamole/instruction.h>
 #include <guacamole/plugin.h>
 #include <guacamole/protocol.h>
+#include <guacamole/socket.h>
 
-#include "client.h"
-#include "log.h"
+#ifdef ENABLE_SSL
+#include <openssl/ssl.h>
+#include "socket-ssl.h"
+#endif
+
+#include <errno.h>
+#include <fcntl.h>
+#include <libgen.h>
+#include <netdb.h>
+#include <netinet/in.h>
+#include <signal.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <syslog.h>
+#include <sys/socket.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <unistd.h>
 
 #define GUACD_DEV_NULL "/dev/null"
 #define GUACD_ROOT     "/"
 
-void guacd_handle_connection(guac_socket* socket) {
+/**
+ * Logs a reasonable explanatory message regarding handshake failure based on
+ * the current value of guac_error.
+ */
+static void guacd_log_handshake_failure() {
+
+    if (guac_error == GUAC_STATUS_CLOSED)
+        guacd_log(GUAC_LOG_INFO,
+                "Guacamole connection closed during handshake");
+    else if (guac_error == GUAC_STATUS_PROTOCOL_ERROR)
+        guacd_log(GUAC_LOG_ERROR,
+                "Guacamole protocol violation. Perhaps the version of "
+                "guacamole-client is incompatible with this version of "
+                "guacd?");
+    else
+        guacd_log(GUAC_LOG_WARNING,
+                "Guacamole handshake failed: %s",
+                guac_status_string(guac_error));
+
+}
+
+/**
+ * Creates a new guac_client for the connection on the given socket, adding
+ * it to the client map based on its ID.
+ */
+static void guacd_handle_connection(guacd_client_map* map, guac_socket* socket) {
 
     guac_client* client;
     guac_client_plugin* plugin;
@@ -83,24 +94,31 @@ void guacd_handle_connection(guac_socket* socket) {
     guac_instruction* connect;
     int init_result;
 
+    /* Reset guac_error */
+    guac_error = GUAC_STATUS_SUCCESS;
+    guac_error_message = NULL;
+
     /* Get protocol from select instruction */
-    select = guac_instruction_expect(
-            socket, GUACD_USEC_TIMEOUT, "select");
+    select = guac_instruction_expect(socket, GUACD_USEC_TIMEOUT, "select");
     if (select == NULL) {
 
         /* Log error */
-        guacd_log_guac_error("Error reading \"select\"");
+        guacd_log_handshake_failure();
+        guacd_log_guac_error(GUAC_LOG_DEBUG,
+                "Error reading \"select\"");
 
         /* Free resources */
         guac_socket_free(socket);
         return;
+
     }
 
     /* Validate args to select */
     if (select->argc != 1) {
 
         /* Log error */
-        guacd_log_error("Bad number of arguments to \"select\" (%i)",
+        guacd_log_handshake_failure();
+        guacd_log(GUAC_LOG_ERROR, "Bad number of arguments to \"select\" (%i)",
                 select->argc);
 
         /* Free resources */
@@ -108,7 +126,7 @@ void guacd_handle_connection(guac_socket* socket) {
         return;
     }
 
-    guacd_log_info("Protocol \"%s\" selected", select->argv[0]);
+    guacd_log(GUAC_LOG_INFO, "Protocol \"%s\" selected", select->argv[0]);
 
     /* Get plugin from protocol in select */
     plugin = guac_client_plugin_open(select->argv[0]);
@@ -117,7 +135,12 @@ void guacd_handle_connection(guac_socket* socket) {
     if (plugin == NULL) {
 
         /* Log error */
-        guacd_log_guac_error("Error loading client plugin");
+        if (guac_error == GUAC_STATUS_NOT_FOUND)
+            guacd_log(GUAC_LOG_WARNING,
+                    "Support for selected protocol is not installed");
+        else
+            guacd_log_guac_error(GUAC_LOG_ERROR,
+                    "Unable to load client plugin");
 
         /* Free resources */
         guac_socket_free(socket);
@@ -129,10 +152,12 @@ void guacd_handle_connection(guac_socket* socket) {
             || guac_socket_flush(socket)) {
 
         /* Log error */
-        guacd_log_guac_error("Error sending \"args\"");
+        guacd_log_handshake_failure();
+        guacd_log_guac_error(GUAC_LOG_DEBUG, "Error sending \"args\"");
 
         if (guac_client_plugin_close(plugin))
-            guacd_log_guac_error("Error closing client plugin");
+            guacd_log_guac_error(GUAC_LOG_WARNING,
+                    "Unable to close client plugin");
 
         guac_socket_free(socket);
         return;
@@ -144,7 +169,8 @@ void guacd_handle_connection(guac_socket* socket) {
     if (size == NULL) {
 
         /* Log error */
-        guacd_log_guac_error("Error reading \"size\"");
+        guacd_log_handshake_failure();
+        guacd_log_guac_error(GUAC_LOG_DEBUG, "Error reading \"size\"");
 
         /* Free resources */
         guac_socket_free(socket);
@@ -157,7 +183,8 @@ void guacd_handle_connection(guac_socket* socket) {
     if (audio == NULL) {
 
         /* Log error */
-        guacd_log_guac_error("Error reading \"audio\"");
+        guacd_log_handshake_failure();
+        guacd_log_guac_error(GUAC_LOG_DEBUG, "Error reading \"audio\"");
 
         /* Free resources */
         guac_socket_free(socket);
@@ -170,7 +197,8 @@ void guacd_handle_connection(guac_socket* socket) {
     if (video == NULL) {
 
         /* Log error */
-        guacd_log_guac_error("Error reading \"video\"");
+        guacd_log_handshake_failure();
+        guacd_log_guac_error(GUAC_LOG_DEBUG, "Error reading \"video\"");
 
         /* Free resources */
         guac_socket_free(socket);
@@ -183,10 +211,12 @@ void guacd_handle_connection(guac_socket* socket) {
     if (connect == NULL) {
 
         /* Log error */
-        guacd_log_guac_error("Error reading \"connect\"");
+        guacd_log_handshake_failure();
+        guacd_log_guac_error(GUAC_LOG_DEBUG, "Error reading \"connect\"");
 
         if (guac_client_plugin_close(plugin))
-            guacd_log_guac_error("Error closing client plugin");
+            guacd_log_guac_error(GUAC_LOG_WARNING,
+                    "Unable to close client plugin");
 
         guac_socket_free(socket);
         return;
@@ -194,14 +224,27 @@ void guacd_handle_connection(guac_socket* socket) {
 
     /* Get client */
     client = guac_client_alloc();
+    if (client == NULL) {
+        guacd_log_guac_error(GUAC_LOG_ERROR, "Unable to create client");
+        guac_socket_free(socket);
+        return;
+    }
+
     client->socket = socket;
-    client->log_info_handler = guacd_client_log_info;
-    client->log_error_handler = guacd_client_log_error;
+    client->log_handler = guacd_client_log;
 
     /* Parse optimal screen dimensions from size instruction */
     client->info.optimal_width  = atoi(size->argv[0]);
     client->info.optimal_height = atoi(size->argv[1]);
 
+    /* If DPI given, set the client resolution */
+    if (size->argc >= 3)
+        client->info.optimal_resolution = atoi(size->argv[2]);
+
+    /* Otherwise, use a safe default for rough backwards compatibility */
+    else
+        client->info.optimal_resolution = 96;
+
     /* Store audio mimetypes */
     client->info.audio_mimetypes = malloc(sizeof(char*) * (audio->argc+1));
     memcpy(client->info.audio_mimetypes, audio->argv,
@@ -214,6 +257,14 @@ void guacd_handle_connection(guac_socket* socket) {
             sizeof(char*) * video->argc);
     client->info.video_mimetypes[video->argc] = NULL;
 
+    /* Store client */
+    if (guacd_client_map_add(map, client))
+        guacd_log(GUAC_LOG_ERROR, "Unable to add client. Internal client storage has failed");
+
+    /* Send connection ID */
+    guacd_log(GUAC_LOG_INFO, "Connection ID is \"%s\"", client->connection_id);
+    guac_protocol_send_ready(socket, client->connection_id);
+
     /* Init client */
     init_result = guac_client_plugin_init_client(plugin,
                 client, connect->argc, connect->argv);
@@ -225,21 +276,26 @@ void guacd_handle_connection(guac_socket* socket) {
 
         guac_client_free(client);
 
-        guacd_log_guac_error("Error instantiating client");
+        guacd_log_guac_error(GUAC_LOG_INFO, "Connection did not succeed");
 
         if (guac_client_plugin_close(plugin))
-            guacd_log_guac_error("Error closing client plugin");
+            guacd_log_guac_error(GUAC_LOG_WARNING,
+                    "Unable to close client plugin");
 
         guac_socket_free(socket);
         return;
     }
 
     /* Start client threads */
-    guacd_log_info("Starting client");
+    guacd_log(GUAC_LOG_INFO, "Starting client");
     if (guacd_client_start(client))
-        guacd_log_error("Client finished abnormally");
+        guacd_log(GUAC_LOG_WARNING, "Client finished abnormally");
     else
-        guacd_log_info("Client finished normally");
+        guacd_log(GUAC_LOG_INFO, "Client disconnected");
+
+    /* Remove client */
+    if (guacd_client_map_remove(map, client->connection_id) == NULL)
+        guacd_log(GUAC_LOG_ERROR, "Unable to remove client. Internal client storage has failed");
 
     /* Free mimetype lists */
     free(client->info.audio_mimetypes);
@@ -253,7 +309,8 @@ void guacd_handle_connection(guac_socket* socket) {
     /* Clean up */
     guac_client_free(client);
     if (guac_client_plugin_close(plugin))
-        guacd_log_error("Error closing client plugin");
+        guacd_log_guac_error(GUAC_LOG_WARNING,
+                "Unable to close client plugin");
 
     /* Close socket */
     guac_socket_free(socket);
@@ -284,13 +341,13 @@ int daemonize() {
     /* Fork once to ensure we aren't the process group leader */
     pid = fork();
     if (pid < 0) {
-        guacd_log_error("Could not fork() parent: %s", strerror(errno));
+        guacd_log(GUAC_LOG_ERROR, "Could not fork() parent: %s", strerror(errno));
         return 1;
     }
 
     /* Exit if we are the parent */
     if (pid > 0) {
-        guacd_log_info("Exiting and passing control to PID %i", pid);
+        guacd_log(GUAC_LOG_DEBUG, "Exiting and passing control to PID %i", pid);
         _exit(0);
     }
 
@@ -300,19 +357,19 @@ int daemonize() {
     /* Fork again so the session group leader exits */
     pid = fork();
     if (pid < 0) {
-        guacd_log_error("Could not fork() group leader: %s", strerror(errno));
+        guacd_log(GUAC_LOG_ERROR, "Could not fork() group leader: %s", strerror(errno));
         return 1;
     }
 
     /* Exit if we are the parent */
     if (pid > 0) {
-        guacd_log_info("Exiting and passing control to PID %i", pid);
+        guacd_log(GUAC_LOG_DEBUG, "Exiting and passing control to PID %i", pid);
         _exit(0);
     }
 
     /* Change to root directory */
     if (chdir(GUACD_ROOT) < 0) {
-        guacd_log_error(
+        guacd_log(GUAC_LOG_ERROR, 
                 "Unable to change working directory to "
                 GUACD_ROOT);
         return 1;
@@ -324,7 +381,7 @@ int daemonize() {
     || redirect_fd(STDOUT_FILENO, O_WRONLY)
     || redirect_fd(STDERR_FILENO, O_WRONLY)) {
 
-        guacd_log_error(
+        guacd_log(GUAC_LOG_ERROR, 
                 "Unable to redirect standard file descriptors to "
                 GUACD_DEV_NULL);
         return 1;
@@ -335,7 +392,6 @@ int daemonize() {
 
 }
 
-
 int main(int argc, char* argv[]) {
 
     /* Server */
@@ -357,85 +413,32 @@ int main(int argc, char* argv[]) {
     socklen_t client_addr_len;
     int connected_socket_fd;
 
-    /* Arguments */
-    char* listen_address = NULL; /* Default address of INADDR_ANY */
-    char* listen_port = "4822";  /* Default port */
-    char* pidfile = NULL;
-    int opt;
-    int foreground = 0;
-
 #ifdef ENABLE_SSL
-    /* SSL */
-    char* cert_file = NULL;
-    char* key_file = NULL;
     SSL_CTX* ssl_context = NULL;
 #endif
 
+    guacd_client_map* map = guacd_client_map_alloc();
+
     /* General */
     int retval;
 
-    /* Parse arguments */
-    while ((opt = getopt(argc, argv, "l:b:p:C:K:f")) != -1) {
-        if (opt == 'l') {
-            listen_port = strdup(optarg);
-        }
-        else if (opt == 'b') {
-            listen_address = strdup(optarg);
-        }
-        else if (opt == 'f') {
-            foreground = 1;
-        }
-        else if (opt == 'p') {
-            pidfile = strdup(optarg);
-        }
-#ifdef ENABLE_SSL
-        else if (opt == 'C') {
-            cert_file = strdup(optarg);
-        }
-        else if (opt == 'K') {
-            key_file = strdup(optarg);
-        }
-#else
-        else if (opt == 'C' || opt == 'K') {
-            fprintf(stderr,
-                    "This guacd does not have SSL/TLS support compiled in.\n\n"
+    /* Load configuration */
+    guacd_config* config = guacd_conf_load();
+    if (config == NULL || guacd_conf_parse_args(config, argc, argv))
+       exit(EXIT_FAILURE);
 
-                    "If you wish to enable support for the -%c option, please install libssl and\n"
-                    "recompile guacd.\n",
-                    opt);
-            exit(EXIT_FAILURE);
-        }
-#endif
-        else {
-
-            fprintf(stderr, "USAGE: %s"
-                    " [-l LISTENPORT]"
-                    " [-b LISTENADDRESS]"
-                    " [-p PIDFILE]"
-#ifdef ENABLE_SSL
-                    " [-C CERTIFICATE_FILE]"
-                    " [-K PEM_FILE]"
-#endif
-                    " [-f]\n", argv[0]);
-
-            exit(EXIT_FAILURE);
-        }
-    }
-
-    /* Set up logging prefix */
-    strncpy(log_prefix, basename(argv[0]), sizeof(log_prefix));
-
-    /* Open log as early as we can */
-    openlog(NULL, LOG_PID, LOG_DAEMON);
+    /* Init logging as early as possible */
+    guacd_log_level = config->max_log_level;
+    openlog(GUACD_LOG_NAME, LOG_PID, LOG_DAEMON);
 
     /* Log start */
-    guacd_log_info("Guacamole proxy daemon (guacd) version " VERSION);
+    guacd_log(GUAC_LOG_INFO, "Guacamole proxy daemon (guacd) version " VERSION " started");
 
     /* Get addresses for binding */
-    if ((retval = getaddrinfo(listen_address, listen_port,
+    if ((retval = getaddrinfo(config->bind_host, config->bind_port,
                     &hints, &addresses))) {
 
-        guacd_log_error("Error parsing given address or port: %s",
+        guacd_log(GUAC_LOG_ERROR, "Error parsing given address or port: %s",
                 gai_strerror(retval));
         exit(EXIT_FAILURE);
 
@@ -444,14 +447,14 @@ int main(int argc, char* argv[]) {
     /* Get socket */
     socket_fd = socket(AF_INET, SOCK_STREAM, 0);
     if (socket_fd < 0) {
-        guacd_log_error("Error opening socket: %s", strerror(errno));
+        guacd_log(GUAC_LOG_ERROR, "Error opening socket: %s", strerror(errno));
         exit(EXIT_FAILURE);
     }
 
     /* Allow socket reuse */
     if (setsockopt(socket_fd, SOL_SOCKET, SO_REUSEADDR,
                 (void*) &opt_on, sizeof(opt_on))) {
-        guacd_log_info("Unable to set socket options for reuse: %s",
+        guacd_log(GUAC_LOG_WARNING, "Unable to set socket options for reuse: %s",
                 strerror(errno));
     }
 
@@ -467,7 +470,7 @@ int main(int argc, char* argv[]) {
                 bound_address, sizeof(bound_address),
                 bound_port, sizeof(bound_port),
                 NI_NUMERICHOST | NI_NUMERICSERV)))
-            guacd_log_error("Unable to resolve host: %s",
+            guacd_log(GUAC_LOG_ERROR, "Unable to resolve host: %s",
                     gai_strerror(retval));
 
         /* Attempt to bind socket to address */
@@ -475,7 +478,7 @@ int main(int argc, char* argv[]) {
                     current_address->ai_addr,
                     current_address->ai_addrlen) == 0) {
 
-            guacd_log_info("Successfully bound socket to "
+            guacd_log(GUAC_LOG_DEBUG, "Successfully bound socket to "
                     "host %s, port %s", bound_address, bound_port);
 
             /* Done if successful bind */
@@ -485,7 +488,7 @@ int main(int argc, char* argv[]) {
 
         /* Otherwise log information regarding bind failure */
         else
-            guacd_log_info("Unable to bind socket to "
+            guacd_log(GUAC_LOG_DEBUG, "Unable to bind socket to "
                     "host %s, port %s: %s",
                     bound_address, bound_port, strerror(errno));
 
@@ -495,61 +498,61 @@ int main(int argc, char* argv[]) {
 
     /* If unable to bind to anything, fail */
     if (current_address == NULL) {
-        guacd_log_error("Unable to bind socket to any addresses.");
+        guacd_log(GUAC_LOG_ERROR, "Unable to bind socket to any addresses.");
         exit(EXIT_FAILURE);
     }
 
 #ifdef ENABLE_SSL
     /* Init SSL if enabled */
-    if (key_file != NULL || cert_file != NULL) {
+    if (config->key_file != NULL || config->cert_file != NULL) {
 
         /* Init SSL */
-        guacd_log_info("Communication will require SSL/TLS.");
+        guacd_log(GUAC_LOG_INFO, "Communication will require SSL/TLS.");
         SSL_library_init();
         SSL_load_error_strings();
         ssl_context = SSL_CTX_new(SSLv23_server_method());
 
         /* Load key */
-        if (key_file != NULL) {
-            guacd_log_info("Using PEM keyfile %s", key_file);
-            if (!SSL_CTX_use_PrivateKey_file(ssl_context, key_file, SSL_FILETYPE_PEM)) {
-                guacd_log_error("Unable to load keyfile.");
+        if (config->key_file != NULL) {
+            guacd_log(GUAC_LOG_INFO, "Using PEM keyfile %s", config->key_file);
+            if (!SSL_CTX_use_PrivateKey_file(ssl_context, config->key_file, SSL_FILETYPE_PEM)) {
+                guacd_log(GUAC_LOG_ERROR, "Unable to load keyfile.");
                 exit(EXIT_FAILURE);
             }
         }
         else
-            guacd_log_info("No PEM keyfile given - SSL/TLS may not work.");
+            guacd_log(GUAC_LOG_WARNING, "No PEM keyfile given - SSL/TLS may not work.");
 
         /* Load cert file if specified */
-        if (cert_file != NULL) {
-            guacd_log_info("Using certificate file %s", cert_file);
-            if (!SSL_CTX_use_certificate_file(ssl_context, cert_file, SSL_FILETYPE_PEM)) {
-                guacd_log_error("Unable to load certificate.");
+        if (config->cert_file != NULL) {
+            guacd_log(GUAC_LOG_INFO, "Using certificate file %s", config->cert_file);
+            if (!SSL_CTX_use_certificate_chain_file(ssl_context, config->cert_file)) {
+                guacd_log(GUAC_LOG_ERROR, "Unable to load certificate.");
                 exit(EXIT_FAILURE);
             }
         }
         else
-            guacd_log_info("No certificate file given - SSL/TLS may not work.");
+            guacd_log(GUAC_LOG_WARNING, "No certificate file given - SSL/TLS may not work.");
 
     }
 #endif
 
     /* Daemonize if requested */
-    if (!foreground) {
+    if (!config->foreground) {
 
         /* Attempt to daemonize process */
         if (daemonize()) {
-            guacd_log_error("Could not become a daemon.");
+            guacd_log(GUAC_LOG_ERROR, "Could not become a daemon.");
             exit(EXIT_FAILURE);
         }
 
     }
 
     /* Write PID file if requested */
-    if (pidfile != NULL) {
+    if (config->pidfile != NULL) {
 
         /* Attempt to open pidfile and write PID */
-        FILE* pidf = fopen(pidfile, "w");
+        FILE* pidf = fopen(config->pidfile, "w");
         if (pidf) {
             fprintf(pidf, "%d\n", getpid());
             fclose(pidf);
@@ -557,7 +560,7 @@ int main(int argc, char* argv[]) {
         
         /* Fail if could not write PID file*/
         else {
-            guacd_log_error("Could not write PID file: %s", strerror(errno));
+            guacd_log(GUAC_LOG_ERROR, "Could not write PID file: %s", strerror(errno));
             exit(EXIT_FAILURE);
         }
 
@@ -565,18 +568,18 @@ int main(int argc, char* argv[]) {
 
     /* Ignore SIGPIPE */
     if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
-        guacd_log_info("Could not set handler for SIGPIPE to ignore. "
+        guacd_log(GUAC_LOG_INFO, "Could not set handler for SIGPIPE to ignore. "
                 "SIGPIPE may cause termination of the daemon.");
     }
 
     /* Ignore SIGCHLD (force automatic removal of children) */
     if (signal(SIGCHLD, SIG_IGN) == SIG_ERR) {
-        guacd_log_info("Could not set handler for SIGCHLD to ignore. "
+        guacd_log(GUAC_LOG_INFO, "Could not set handler for SIGCHLD to ignore. "
                 "Child processes may pile up in the process table.");
     }
 
     /* Log listening status */
-    guacd_log_info("Listening on host %s, port %s", bound_address, bound_port);
+    guacd_log(GUAC_LOG_INFO, "Listening on host %s, port %s", bound_address, bound_port);
 
     /* Free addresses */
     freeaddrinfo(addresses);
@@ -588,7 +591,7 @@ int main(int argc, char* argv[]) {
 
         /* Listen for connections */
         if (listen(socket_fd, 5) < 0) {
-            guacd_log_error("Could not listen on socket: %s", strerror(errno));
+            guacd_log(GUAC_LOG_ERROR, "Could not listen on socket: %s", strerror(errno));
             return 3;
         }
 
@@ -598,7 +601,7 @@ int main(int argc, char* argv[]) {
                 (struct sockaddr*) &client_addr, &client_addr_len);
 
         if (connected_socket_fd < 0) {
-            guacd_log_error("Could not accept client connection: %s",
+            guacd_log(GUAC_LOG_ERROR, "Could not accept client connection: %s",
                     strerror(errno));
             return 3;
         }
@@ -616,7 +619,7 @@ int main(int argc, char* argv[]) {
 
         /* If error, log */
         if (child_pid == -1)
-            guacd_log_error("Error forking child process: %s", strerror(errno));
+            guacd_log(GUAC_LOG_ERROR, "Error forking child process: %s", strerror(errno));
 
         /* If child, start client, and exit when finished */
         else if (child_pid == 0) {
@@ -629,7 +632,8 @@ int main(int argc, char* argv[]) {
             if (ssl_context != NULL) {
                 socket = guac_socket_open_secure(ssl_context, connected_socket_fd);
                 if (socket == NULL) {
-                    guacd_log_guac_error("Error opening secure connection");
+                    guacd_log_guac_error(GUAC_LOG_ERROR,
+                            "Unable to set up SSL/TLS");
                     return 0;
                 }
             }
@@ -640,14 +644,14 @@ int main(int argc, char* argv[]) {
             socket = guac_socket_open(connected_socket_fd);
 #endif
 
-            guacd_handle_connection(socket);
+            guacd_handle_connection(map, socket);
             close(connected_socket_fd);
             return 0;
         }
 
         /* If parent, close reference to child's descriptor */
         else if (close(connected_socket_fd) < 0) {
-            guacd_log_error("Error closing daemon reference to "
+            guacd_log(GUAC_LOG_ERROR, "Error closing daemon reference to "
                     "child descriptor: %s", strerror(errno));
         }
 
@@ -655,7 +659,7 @@ int main(int argc, char* argv[]) {
 
     /* Close socket */
     if (close(socket_fd) < 0) {
-        guacd_log_error("Could not close socket: %s", strerror(errno));
+        guacd_log(GUAC_LOG_ERROR, "Could not close socket: %s", strerror(errno));
         return 3;
     }
 
