@@ -1,180 +1,49 @@
-
-/* ***** BEGIN LICENSE BLOCK *****
- * Version: MPL 1.1/GPL 2.0/LGPL 2.1
+/*
+ * Copyright (C) 2013 Glyptodon LLC
  *
- * The contents of this file are subject to the Mozilla Public License Version
- * 1.1 (the "License"); you may not use this file except in compliance with
- * the License. You may obtain a copy of the License at
- * http://www.mozilla.org/MPL/
+ * Permission is hereby granted, free of charge, to any person obtaining a copy
+ * of this software and associated documentation files (the "Software"), to deal
+ * in the Software without restriction, including without limitation the rights
+ * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+ * copies of the Software, and to permit persons to whom the Software is
+ * furnished to do so, subject to the following conditions:
  *
- * Software distributed under the License is distributed on an "AS IS" basis,
- * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
- * for the specific language governing rights and limitations under the
- * License.
+ * The above copyright notice and this permission notice shall be included in
+ * all copies or substantial portions of the Software.
  *
- * The Original Code is libguac.
- *
- * The Initial Developer of the Original Code is
- * Michael Jumper.
- * Portions created by the Initial Developer are Copyright (C) 2010
- * the Initial Developer. All Rights Reserved.
- *
- * Contributor(s):
- *
- * Alternatively, the contents of this file may be used under the terms of
- * either the GNU General Public License Version 2 or later (the "GPL"), or
- * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
- * in which case the provisions of the GPL or the LGPL are applicable instead
- * of those above. If you wish to allow use of your version of this file only
- * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the MPL, indicate your
- * decision by deleting the provisions above and replace them with the notice
- * and other provisions required by the GPL or the LGPL. If you do not delete
- * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the MPL, the GPL or the LGPL.
- *
- * ***** END LICENSE BLOCK ***** */
-
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+ * THE SOFTWARE.
+ */
 
 #ifndef _GUAC_CLIENT_H
 #define _GUAC_CLIENT_H
 
-#include <stdarg.h>
-
-#include "instruction.h"
-#include "layer.h"
-#include "pool.h"
-#include "socket.h"
-#include "stream.h"
-#include "timestamp.h"
-
 /**
- * Provides functions and structures required for defining (and handling) a proxy client.
+ * Functions and structure contents for the Guacamole proxy client.
  *
  * @file client.h
  */
 
-typedef struct guac_client guac_client;
+#include "client-fntypes.h"
+#include "client-types.h"
+#include "client-constants.h"
+#include "instruction-types.h"
+#include "layer-types.h"
+#include "object-types.h"
+#include "pool-types.h"
+#include "socket-types.h"
+#include "stream-types.h"
+#include "timestamp-types.h"
 
-/**
- * Handler for server messages (where "server" refers to the server that
- * the proxy client is connected to).
- */
-typedef int guac_client_handle_messages(guac_client* client);
+#include <cairo/cairo.h>
+#include <stdarg.h>
 
-/**
- * Handler for Guacamole mouse events.
- */
-typedef int guac_client_mouse_handler(guac_client* client, int x, int y, int button_mask);
-
-/**
- * Handler for Guacamole key events.
- */
-typedef int guac_client_key_handler(guac_client* client, int keysym, int pressed);
-
-/**
- * Handler for Guacamole clipboard events.
- */
-typedef int guac_client_clipboard_handler(guac_client* client, char* copied);
-
-/**
- * Handler for Guacamole screen size events.
- */
-typedef int guac_client_size_handler(guac_client* client,
-        int width, int height);
-
-/**
- * Handler for Guacamole audio format events.
- */
-typedef int guac_client_audio_handler(guac_client* client, char* mimetype);
-
-/**
- * Handler for Guacamole video format events.
- */
-typedef int guac_client_video_handler(guac_client* client, char* mimetype);
-
-/**
- * Handler for freeing up any extra data allocated by the client
- * implementation.
- */
-typedef int guac_client_free_handler(guac_client* client);
-
-/**
- * Handler for logging messages
- */
-typedef void guac_client_log_handler(guac_client* client, const char* format, va_list args); 
-
-/**
- * Handler which should initialize the given guac_client.
- */
-typedef int guac_client_init_handler(guac_client* client, int argc, char** argv);
-
-/**
- * The flag set in the mouse button mask when the left mouse button is down.
- */
-#define GUAC_CLIENT_MOUSE_LEFT        0x01
-
-/**
- * The flag set in the mouse button mask when the middle mouse button is down.
- */
-#define GUAC_CLIENT_MOUSE_MIDDLE      0x02
-
-/**
- * The flag set in the mouse button mask when the right mouse button is down.
- */
-#define GUAC_CLIENT_MOUSE_RIGHT       0x04
-
-/**
- * The flag set in the mouse button mask when the mouse scrollwheel is scrolled
- * up. Note that mouse scrollwheels are actually sets of two buttons. One
- * button is pressed and released for an upward scroll, and the other is
- * pressed and released for a downward scroll. Some mice may actually implement
- * these as separate buttons, not a wheel.
- */
-#define GUAC_CLIENT_MOUSE_SCROLL_UP   0x08
-
-/**
- * The flag set in the mouse button mask when the mouse scrollwheel is scrolled
- * down. Note that mouse scrollwheels are actually sets of two buttons. One
- * button is pressed and released for an upward scroll, and the other is
- * pressed and released for a downward scroll. Some mice may actually implement
- * these as separate buttons, not a wheel.
- */
-#define GUAC_CLIENT_MOUSE_SCROLL_DOWN 0x10
-
-/**
- * The minimum number of buffers to create before allowing free'd buffers to
- * be reclaimed. In the case a protocol rapidly creates, uses, and destroys
- * buffers, this can prevent unnecessary reuse of the same buffer (which
- * would make draw operations unnecessarily synchronous).
- */
-#define GUAC_BUFFER_POOL_INITIAL_SIZE 1024
-
-/**
- * Possible current states of the Guacamole client. Currently, the only
- * two states are GUAC_CLIENT_RUNNING and GUAC_CLIENT_STOPPING.
- */
-typedef enum guac_client_state {
-
-    /**
-     * The state of the client from when it has been allocated by the main
-     * daemon until it is killed or disconnected.
-     */
-    GUAC_CLIENT_RUNNING,
-
-    /**
-     * The state of the client when a stop has been requested, signalling the
-     * I/O threads to shutdown.
-     */
-    GUAC_CLIENT_STOPPING
-
-} guac_client_state;
-
-/**
- * Information exposed by the remote client during the connection handshake
- * which can be used by a client plugin.
- */
-typedef struct guac_client_info {
+struct guac_client_info {
 
     /**
      * The number of pixels the remote client requests for the display width.
@@ -204,14 +73,17 @@ typedef struct guac_client_info {
      */
     const char** video_mimetypes;
 
-} guac_client_info;
+    /**
+     * The DPI of the physical remote display if configured for the optimal
+     * width/height combination described here. This need not be honored by
+     * a client plugin implementation, but if the underlying protocol of the
+     * client plugin supports dynamic sizing of the screen, honoring the
+     * stated resolution of the display size request is recommended.
+     */
+    int optimal_resolution;
+
+};
 
-/**
- * Guacamole proxy client.
- *
- * Represents a Guacamole proxy client (the client which communicates to
- * a server on behalf of Guacamole, on behalf of the web-client).
- */
 struct guac_client {
 
     /**
@@ -322,14 +194,14 @@ struct guac_client {
      * handler will be called whenever the web-client sets the data of the
      * clipboard.
      *
-     * This handler takes a single string which contains the text which
-     * has been set in the clipboard. This text is already unescaped from
-     * the Guacamole escaped version sent within the clipboard message
-     * in the protocol.
+     * The handler takes a guac_stream, which contains the stream index and
+     * will persist through the duration of the transfer, and the mimetype
+     * of the data being transferred.
      *
      * Example:
      * @code
-     *     int clipboard_handler(guac_client* client, char* copied);
+     *     int clipboard_handler(guac_client* client, guac_stream* stream,
+     *             char* mimetype);
      *
      *     int guac_client_init(guac_client* client, int argc, char** argv) {
      *         client->clipboard_handler = clipboard_handler;
@@ -355,6 +227,100 @@ struct guac_client {
      */
     guac_client_size_handler* size_handler;
 
+    /**
+     * Handler for file events sent by the Guacamole web-client.
+     *
+     * The handler takes a guac_stream which contains the stream index and
+     * will persist through the duration of the transfer, the mimetype of
+     * the file being transferred, and the filename.
+     *
+     * Example:
+     * @code
+     *     int file_handler(guac_client* client, guac_stream* stream,
+     *             char* mimetype, char* filename);
+     *
+     *     int guac_client_init(guac_client* client, int argc, char** argv) {
+     *         client->file_handler = file_handler;
+     *     }
+     * @endcode
+     */
+    guac_client_file_handler* file_handler;
+
+    /**
+     * Handler for pipe events sent by the Guacamole web-client.
+     *
+     * The handler takes a guac_stream which contains the stream index and
+     * will persist through the duration of the transfer, the mimetype of
+     * the data being transferred, and the pipe name.
+     *
+     * Example:
+     * @code
+     *     int pipe_handler(guac_client* client, guac_stream* stream,
+     *             char* mimetype, char* name);
+     *
+     *     int guac_client_init(guac_client* client, int argc, char** argv) {
+     *         client->pipe_handler = pipe_handler;
+     *     }
+     * @endcode
+     */
+    guac_client_pipe_handler* pipe_handler;
+
+    /**
+     * Handler for ack events sent by the Guacamole web-client.
+     *
+     * The handler takes a guac_stream which contains the stream index and
+     * will persist through the duration of the transfer, a string containing
+     * the error or status message, and a status code.
+     *
+     * Example:
+     * @code
+     *     int ack_handler(guac_client* client, guac_stream* stream,
+     *             char* error, guac_protocol_status status);
+     *
+     *     int guac_client_init(guac_client* client, int argc, char** argv) {
+     *         client->ack_handler = ack_handler;
+     *     }
+     * @endcode
+     */
+    guac_client_ack_handler* ack_handler;
+
+    /**
+     * Handler for blob events sent by the Guacamole web-client.
+     *
+     * The handler takes a guac_stream which contains the stream index and
+     * will persist through the duration of the transfer, an arbitrary buffer
+     * containing the blob, and the length of the blob.
+     *
+     * Example:
+     * @code
+     *     int blob_handler(guac_client* client, guac_stream* stream,
+     *             void* data, int length);
+     *
+     *     int guac_client_init(guac_client* client, int argc, char** argv) {
+     *         client->blob_handler = blob_handler;
+     *     }
+     * @endcode
+     */
+    guac_client_blob_handler* blob_handler;
+
+    /**
+     * Handler for stream end events sent by the Guacamole web-client.
+     *
+     * The handler takes only a guac_stream which contains the stream index.
+     * This guac_stream will be disposed of immediately after this event is
+     * finished.
+     *
+     * Example:
+     * @code
+     *     int end_handler(guac_client* client, guac_stream* stream);
+     *
+     *     int guac_client_init(guac_client* client, int argc, char** argv) {
+     *         client->end_handler = end_handler;
+     *     }
+     * @endcode
+     */
+    guac_client_end_handler* end_handler;
+
     /**
      * Handler for freeing data when the client is being unloaded.
      *
@@ -379,8 +345,8 @@ struct guac_client {
     guac_client_free_handler* free_handler;
 
     /**
-     * Handler for logging informational messages. This handler will be called
-     * via guac_client_log_info() when the client needs to log information.
+     * Logging handler. This handler will be called via guac_client_log() when
+     * the client needs to log messages of any type.
      *
      * In general, only programs loading the client should implement this
      * handler, as those are the programs that would provide the logging
@@ -391,7 +357,7 @@ struct guac_client {
      *
      * Example:
      * @code
-     *     void log_handler(guac_client* client, const char* format, va_list args);
+     *     void log_handler(guac_client* client, guac_client_log_level level, const char* format, va_list args);
      *
      *     void function_of_daemon() {
      *
@@ -400,32 +366,47 @@ struct guac_client {
      *     }
      * @endcode
      */
-    guac_client_log_handler* log_info_handler;
-
+    guac_client_log_handler* log_handler;
 
     /**
-     * Handler for logging error messages. This handler will be called
-     * via guac_client_log_error() when the client needs to log an error.
+     * Handler for get events sent by the Guacamole web-client.
      *
-     * In general, only programs loading the client should implement this
-     * handler, as those are the programs that would provide the logging
-     * facilities.
-     *
-     * Client implementations should expect these handlers to already be
-     * set.
+     * The handler takes a guac_object, containing the object index which will
+     * persist through the duration of the transfer, and the name of the stream
+     * being requested. It is up to the get handler to create the required body
+     * stream.
      *
      * Example:
      * @code
-     *     void log_handler(guac_client* client, const char* format, va_list args);
-     *
-     *     void function_of_daemon() {
-     *
-     *         guac_client* client = [pass log_handler to guac_client_plugin_get_client()];
+     *     int get_handler(guac_client* client, guac_object* object,
+     *             char* name);
      *
+     *     int guac_client_init(guac_client* client, int argc, char** argv) {
+     *         client->get_handler = get_handler;
      *     }
      * @endcode
      */
-    guac_client_log_handler* log_error_handler;
+    guac_client_get_handler* get_handler;
+
+    /**
+     * Handler for put events sent by the Guacamole web-client.
+     *
+     * The handler takes a guac_object and guac_stream, which each contain their
+     * respective indices which will persist through the duration of the
+     * transfer, the mimetype of the data being transferred, and the name of
+     * the stream within the object being written to.
+     *
+     * Example:
+     * @code
+     *     int put_handler(guac_client* client, guac_object* object,
+     *             guac_stream* stream, char* mimetype, char* name);
+     *
+     *     int guac_client_init(guac_client* client, int argc, char** argv) {
+     *         client->put_handler = put_handler;
+     *     }
+     * @endcode
+     */
+    guac_client_put_handler* put_handler;
 
     /**
      * Pool of buffer indices. Buffers are simply layers with negative indices.
@@ -446,6 +427,35 @@ struct guac_client {
      */
     guac_pool* __stream_pool;
 
+    /**
+     * All available output streams (data going to connected client).
+     */
+    guac_stream* __output_streams;
+
+    /**
+     * All available input streams (data coming from connected client).
+     */
+    guac_stream* __input_streams;
+
+    /**
+     * Pool of object indices.
+     */
+    guac_pool* __object_pool;
+
+    /**
+     * All available objects (arbitrary sets of named streams).
+     */
+    guac_object* __objects;
+
+    /**
+     * The unique identifier allocated for the connection, which may
+     * be used within the Guacamole protocol to refer to this connection.
+     * This identifier is guaranteed to be unique from all existing
+     * connections and will not collide with any available protocol
+     * names.
+     */
+    char* connection_id;
+
 };
 
 /**
@@ -476,54 +486,31 @@ void guac_client_free(guac_client* client);
 int guac_client_handle_instruction(guac_client* client, guac_instruction* instruction);
 
 /**
- * Logs an informational message in the log used by the given client. The
- * logger used will normally be defined by guacd (or whichever program loads
- * the proxy client) by setting the logging handlers of the client when it is
- * loaded.
+ * Writes a message in the log used by the given client. The logger used will
+ * normally be defined by guacd (or whichever program loads the proxy client)
+ * by setting the logging handlers of the client when it is loaded.
  *
- * @param client The proxy client to log an informational message for.
+ * @param client The proxy client logging this message.
+ * @param level The level at which to log this message.
  * @param format A printf-style format string to log.
  * @param ... Arguments to use when filling the format string for printing.
  */
-void guac_client_log_info(guac_client* client, const char* format, ...);
+void guac_client_log(guac_client* client, guac_client_log_level level,
+        const char* format, ...);
 
 /**
- * Logs an error message in the log used by the given client. The logger
- * used will normally be defined by guacd (or whichever program loads the
- * proxy client) by setting the logging handlers of the client when it is
- * loaded.
+ * Writes a message in the log used by the given client. The logger used will
+ * normally be defined by guacd (or whichever program loads the proxy client)
+ * by setting the logging handlers of the client when it is loaded.
  *
- * @param client The proxy client to log an error for.
- * @param format A printf-style format string to log.
- * @param ... Arguments to use when filling the format string for printing.
- */
-void guac_client_log_error(guac_client* client, const char* format, ...);
-
-/**
- * Logs an informational message in the log used by the given client. The
- * logger used will normally be defined by guacd (or whichever program loads
- * the proxy client) by setting the logging handlers of the client when it is
- * loaded.
- *
- * @param client The proxy client to log an informational message for.
+ * @param client The proxy client logging this message.
+ * @param level The level at which to log this message.
  * @param format A printf-style format string to log.
  * @param ap The va_list containing the arguments to be used when filling the
  *           format string for printing.
  */
-void vguac_client_log_info(guac_client* client, const char* format, va_list ap);
-
-/**
- * Logs an error message in the log used by the given client. The logger
- * used will normally be defined by guacd (or whichever program loads the
- * proxy client) by setting the logging handlers of the client when it is
- * loaded.
- *
- * @param client The proxy client to log an error for.
- * @param format A printf-style format string to log.
- * @param ap The va_list containing the arguments to be used when filling the
- *           format string for printing.
- */
-void vguac_client_log_error(guac_client* client, const char* format, va_list ap);
+void vguac_client_log(guac_client* client, guac_client_log_level level,
+        const char* format, va_list ap);
 
 /**
  * Signals the given client to stop gracefully. This is a completely
@@ -534,6 +521,35 @@ void vguac_client_log_error(guac_client* client, const char* format, va_list ap)
  */
 void guac_client_stop(guac_client* client);
 
+/**
+ * Signals the given client to stop gracefully, while also signalling via the
+ * Guacamole protocol that an error has occurred. Note that this is a completely
+ * cooperative signal, and can be ignored by the client or the hosting
+ * daemon. The message given will be logged to the system logs.
+ *
+ * @param client The proxy client to signal to stop.
+ * @param status The status to send over the Guacamole protocol.
+ * @param format A printf-style format string to log.
+ * @param ... Arguments to use when filling the format string for printing.
+ */
+void guac_client_abort(guac_client* client, guac_protocol_status status,
+        const char* format, ...);
+
+/**
+ * Signals the given client to stop gracefully, while also signalling via the
+ * Guacamole protocol that an error has occurred. Note that this is a completely
+ * cooperative signal, and can be ignored by the client or the hosting
+ * daemon. The message given will be logged to the system logs.
+ *
+ * @param client The proxy client to signal to stop.
+ * @param status The status to send over the Guacamole protocol.
+ * @param format A printf-style format string to log.
+ * @param ap The va_list containing the arguments to be used when filling the
+ *           format string for printing.
+ */
+void vguac_client_abort(guac_client* client, guac_protocol_status status,
+        const char* format, va_list ap);
+
 /**
  * Allocates a new buffer (invisible layer). An arbitrary index is
  * automatically assigned if no existing buffer is available for use.
@@ -588,6 +604,95 @@ guac_stream* guac_client_alloc_stream(guac_client* client);
  */
 void guac_client_free_stream(guac_client* client, guac_stream* stream);
 
+/**
+ * Allocates a new object. An arbitrary index is automatically assigned
+ * if no previously-allocated object is available for use.
+ *
+ * @param client
+ *     The proxy client to allocate the object for.
+ *
+ * @return
+ *     The next available object, or a newly allocated object.
+ */
+guac_object* guac_client_alloc_object(guac_client* client);
+
+/**
+ * Returns the given object to the pool of available objects, such that it
+ * can be reused by any subsequent call to guac_client_alloc_object().
+ *
+ * @param client
+ *     The proxy client to return the object to.
+ *
+ * @param object
+ *     The object to return to the pool of available object.
+ */
+void guac_client_free_object(guac_client* client, guac_object* object);
+
+/**
+ * Streams the image data of the given surface over an image stream ("img"
+ * instruction) as PNG-encoded data. The image stream will be automatically
+ * allocated and freed.
+ *
+ * @param client
+ *     The Guacamole client from which the image stream should be allocated.
+ *
+ * @param socket
+ *     The socket over which instructions associated with the image stream
+ *     should be sent.
+ *
+ * @param mode
+ *     The composite mode to use when rendering the image over the given layer.
+ *
+ * @param layer
+ *     The destination layer.
+ *
+ * @param x
+ *     The X coordinate of the upper-left corner of the destination rectangle
+ *     within the given layer.
+ *
+ * @param y
+ *     The Y coordinate of the upper-left corner of the destination rectangle
+ *     within the given layer.
+ *
+ * @param surface
+ *     A Cairo surface containing the image data to be streamed.
+ */
+void guac_client_stream_png(guac_client* client, guac_socket* socket,
+        guac_composite_mode mode, const guac_layer* layer, int x, int y,
+        cairo_surface_t* surface);
+
+/**
+ * Streams the image data of the given surface over an image stream ("img"
+ * instruction) as JPEG-encoded data at the given quality. The image stream
+ * will be automatically allocated and freed.
+ *
+ * @param client
+ *     The Guacamole client from which the image stream should be allocated.
+ *
+ * @param socket
+ *     The socket over which instructions associated with the image stream
+ *     should be sent.
+ *
+ * @param mode
+ *     The composite mode to use when rendering the image over the given layer.
+ *
+ * @param layer
+ *     The destination layer.
+ *
+ * @param x
+ *     The X coordinate of the upper-left corner of the destination rectangle
+ *     within the given layer.
+ *
+ * @param y
+ *     The Y coordinate of the upper-left corner of the destination rectangle
+ *     within the given layer.
+ *
+ * @param surface
+ *     A Cairo surface containing the image data to be streamed.
+ */
+void guac_client_stream_jpeg(guac_client* client, guac_socket* socket,
+        guac_composite_mode mode, const guac_layer* layer, int x, int y,
+        cairo_surface_t* surface, int quality);
 
 /**
  * The default Guacamole client layer, layer 0.
@@ -595,3 +700,4 @@ void guac_client_free_stream(guac_client* client, guac_stream* stream);
 extern const guac_layer* GUAC_DEFAULT_LAYER;
 
 #endif
+
