@@ -1,49 +1,28 @@
-
-/* ***** BEGIN LICENSE BLOCK *****
- * Version: MPL 1.1/GPL 2.0/LGPL 2.1
+/*
+ * Copyright (C) 2013 Glyptodon LLC
  *
- * The contents of this file are subject to the Mozilla Public License Version
- * 1.1 (the "License"); you may not use this file except in compliance with
- * the License. You may obtain a copy of the License at
- * http://www.mozilla.org/MPL/
+ * Permission is hereby granted, free of charge, to any person obtaining a copy
+ * of this software and associated documentation files (the "Software"), to deal
+ * in the Software without restriction, including without limitation the rights
+ * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+ * copies of the Software, and to permit persons to whom the Software is
+ * furnished to do so, subject to the following conditions:
  *
- * Software distributed under the License is distributed on an "AS IS" basis,
- * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
- * for the specific language governing rights and limitations under the
- * License.
+ * The above copyright notice and this permission notice shall be included in
+ * all copies or substantial portions of the Software.
  *
- * The Original Code is libguac.
- *
- * The Initial Developer of the Original Code is
- * Michael Jumper.
- * Portions created by the Initial Developer are Copyright (C) 2010
- * the Initial Developer. All Rights Reserved.
- *
- * Contributor(s):
- *
- * Alternatively, the contents of this file may be used under the terms of
- * either the GNU General Public License Version 2 or later (the "GPL"), or
- * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
- * in which case the provisions of the GPL or the LGPL are applicable instead
- * of those above. If you wish to allow use of your version of this file only
- * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the MPL, indicate your
- * decision by deleting the provisions above and replace them with the notice
- * and other provisions required by the GPL or the LGPL. If you do not delete
- * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the MPL, the GPL or the LGPL.
- *
- * ***** END LICENSE BLOCK ***** */
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+ * THE SOFTWARE.
+ */
 
 #ifndef _GUAC_PROTOCOL_H
 #define _GUAC_PROTOCOL_H
 
-#include <cairo/cairo.h>
-
-#include "layer.h"
-#include "socket.h"
-#include "timestamp.h"
-
 /**
  * Provides functions and structures required for communicating using the
  * Guacamole protocol over a guac_socket connection, such as that provided by
@@ -52,122 +31,35 @@
  * @file protocol.h
  */
 
-/**
- * Composite modes used by Guacamole draw instructions. Each
- * composite mode maps to a unique channel mask integer.
- */
-typedef enum guac_composite_mode {
+#include "layer-types.h"
+#include "object-types.h"
+#include "protocol-types.h"
+#include "socket-types.h"
+#include "stream-types.h"
+#include "timestamp-types.h"
 
-    /*
-     * A: Source where destination transparent = S n D'
-     * B: Source where destination opaque      = S n D
-     * C: Destination where source transparent = D n S'
-     * D: Destination where source opaque      = D n S
-     *
-     * 0 = Active, 1 = Inactive
-     */
-                           /* ABCD */
-    GUAC_COMP_ROUT  = 0x2, /* 0010 - Clears destination where source opaque  */
-    GUAC_COMP_ATOP  = 0x6, /* 0110 - Fill where destination opaque only      */
-    GUAC_COMP_XOR   = 0xA, /* 1010 - XOR                                     */
-    GUAC_COMP_ROVER = 0xB, /* 1011 - Fill where destination transparent only */
-    GUAC_COMP_OVER  = 0xE, /* 1110 - Draw normally                           */
-    GUAC_COMP_PLUS  = 0xF, /* 1111 - Add                                     */
-
-    /* Unimplemented in client: */
-    /* NOT IMPLEMENTED:       0000 - Clear          */
-    /* NOT IMPLEMENTED:       0011 - No operation   */
-    /* NOT IMPLEMENTED:       0101 - Additive IN    */
-    /* NOT IMPLEMENTED:       0111 - Additive ATOP  */
-    /* NOT IMPLEMENTED:       1101 - Additive RATOP */
-
-    /* Buggy in webkit browsers, as they keep channel C on in all cases: */
-    GUAC_COMP_RIN   = 0x1, /* 0001 */
-    GUAC_COMP_IN    = 0x4, /* 0100 */
-    GUAC_COMP_OUT   = 0x8, /* 1000 */
-    GUAC_COMP_RATOP = 0x9, /* 1001 */
-    GUAC_COMP_SRC   = 0xC  /* 1100 */
-
-    /* Bitwise composite operations (binary) */
-
-    /*
-     * A: S' & D'
-     * B: S' & D
-     * C: S  & D'
-     * D: S  & D
-     *
-     * 0 = Active, 1 = Inactive
-     */
-
-} guac_composite_mode;
-
-
-/**
- * Default transfer functions. There is no current facility in the
- * Guacamole protocol to define custom transfer functions.
- */
-typedef enum guac_transfer_function {
-
-    /* Constant functions */               /* ABCD */
-    GUAC_TRANSFER_BINARY_BLACK      = 0x0, /* 0000 */
-    GUAC_TRANSFER_BINARY_WHITE      = 0xF, /* 1111 */
-
-    /* Copy functions */
-    GUAC_TRANSFER_BINARY_SRC        = 0x3, /* 0011 */
-    GUAC_TRANSFER_BINARY_DEST       = 0x5, /* 0101 */
-    GUAC_TRANSFER_BINARY_NSRC       = 0xC, /* 1100 */
-    GUAC_TRANSFER_BINARY_NDEST      = 0xA, /* 1010 */
-
-    /* AND / NAND */
-    GUAC_TRANSFER_BINARY_AND        = 0x1, /* 0001 */
-    GUAC_TRANSFER_BINARY_NAND       = 0xE, /* 1110 */
-
-    /* OR / NOR */
-    GUAC_TRANSFER_BINARY_OR         = 0x7, /* 0111 */
-    GUAC_TRANSFER_BINARY_NOR        = 0x8, /* 1000 */
-
-    /* XOR / XNOR */
-    GUAC_TRANSFER_BINARY_XOR        = 0x6, /* 0110 */
-    GUAC_TRANSFER_BINARY_XNOR       = 0x9, /* 1001 */
-
-    /* AND / NAND with inverted source */
-    GUAC_TRANSFER_BINARY_NSRC_AND   = 0x4, /* 0100 */
-    GUAC_TRANSFER_BINARY_NSRC_NAND  = 0xB, /* 1011 */
-
-    /* OR / NOR with inverted source */
-    GUAC_TRANSFER_BINARY_NSRC_OR    = 0xD, /* 1101 */
-    GUAC_TRANSFER_BINARY_NSRC_NOR   = 0x2, /* 0010 */
-
-    /* AND / NAND with inverted destination */
-    GUAC_TRANSFER_BINARY_NDEST_AND  = 0x2, /* 0010 */
-    GUAC_TRANSFER_BINARY_NDEST_NAND = 0xD, /* 1101 */
-
-    /* OR / NOR with inverted destination */
-    GUAC_TRANSFER_BINARY_NDEST_OR   = 0xB, /* 1011 */
-    GUAC_TRANSFER_BINARY_NDEST_NOR  = 0x4  /* 0100 */
-
-} guac_transfer_function;
-
-/**
- * Supported line cap styles
- */
-typedef enum guac_line_cap_style {
-    GUAC_LINE_CAP_BUTT   = 0x0,
-    GUAC_LINE_CAP_ROUND  = 0x1,
-    GUAC_LINE_CAP_SQUARE = 0x2
-} guac_line_cap_style;
-
-/**
- * Supported line join styles
- */
-typedef enum guac_line_join_style {
-    GUAC_LINE_JOIN_BEVEL = 0x0,
-    GUAC_LINE_JOIN_MITER = 0x1,
-    GUAC_LINE_JOIN_ROUND = 0x2
-} guac_line_join_style;
+#include <cairo/cairo.h>
+#include <stdarg.h>
 
 /* CONTROL INSTRUCTIONS */
 
+/**
+ * Sends an ack instruction over the given guac_socket connection.
+ *
+ * If an error occurs sending the instruction, a non-zero value is
+ * returned, and guac_error is set appropriately.
+ *
+ * @param socket The guac_socket connection to use.
+ * @param stream The guac_stream associated with the operation this ack is
+ *               acknowledging.
+ * @param error The human-readable description associated with the error or
+ *              status update.
+ * @param status The status code related to the error or status.
+ * @return Zero on success, non-zero on error.
+ */
+int guac_protocol_send_ack(guac_socket* socket, guac_stream* stream,
+        const char* error, guac_protocol_status status);
+
 /**
  * Sends an args instruction over the given guac_socket connection.
  *
@@ -210,10 +102,41 @@ int guac_protocol_send_disconnect(guac_socket* socket);
  * returned, and guac_error is set appropriately.
  *
  * @param socket The guac_socket connection to use.
- * @param error The description associated with the error.
+ * @param error The human-readable description associated with the error.
+ * @param status The status code related to the error.
  * @return Zero on success, non-zero on error.
  */
-int guac_protocol_send_error(guac_socket* socket, const char* error);
+int guac_protocol_send_error(guac_socket* socket, const char* error,
+        guac_protocol_status status);
+
+/**
+ * Sends a log instruction over the given guac_socket connection. This is
+ * mainly useful in debugging.
+ *
+ * If an error occurs sending the instruction, a non-zero value is
+ * returned, and guac_error is set appropriately.
+ *
+ * @param socket The guac_socket connection to use.
+ * @param format A printf-style format string to log.
+ * @param ... Arguments to use when filling the format string for printing.
+ * @return Zero on success, non-zero on error.
+ */
+int guac_protocol_send_log(guac_socket* socket, const char* format, ...);
+
+/**
+ * Sends a log instruction over the given guac_socket connection. This is
+ * mainly useful in debugging.
+ *
+ * If an error occurs sending the instruction, a non-zero value is
+ * returned, and guac_error is set appropriately.
+ *
+ * @param socket The guac_socket connection to use.
+ * @param format A printf-style format string to log.
+ * @param ap The va_list containing the arguments to be used when filling the
+ *           format string for printing.
+ */
+int vguac_protocol_send_log(guac_socket* socket, const char* format,
+        va_list args);
 
 /**
  * Sends a nest instruction over the given guac_socket connection.
@@ -231,6 +154,30 @@ int guac_protocol_send_error(guac_socket* socket, const char* error);
 int guac_protocol_send_nest(guac_socket* socket, int index,
         const char* data);
 
+/**
+ * Sends a nop instruction (null-operation) over the given guac_socket
+ * connection.
+ *
+ * If an error occurs sending the instruction, a non-zero value is
+ * returned, and guac_error is set appropriately.
+ *
+ * @param socket The guac_socket connection to use.
+ * @return Zero on success, non-zero on error.
+ */
+int guac_protocol_send_nop(guac_socket* socket);
+
+/**
+ * Sends a ready instruction over the given guac_socket connection.
+ *
+ * If an error occurs sending the instruction, a non-zero value is
+ * returned, and guac_error is set appropriately.
+ *
+ * @param socket The guac_socket connection to use.
+ * @param id The connection ID of the connection that is ready.
+ * @return Zero on success, non-zero on error.
+ */
+int guac_protocol_send_ready(guac_socket* socket, const char* id);
+
 /**
  * Sends a set instruction over the given guac_socket connection.
  *
@@ -271,6 +218,75 @@ int guac_protocol_send_select(guac_socket* socket, const char* protocol);
  */
 int guac_protocol_send_sync(guac_socket* socket, guac_timestamp timestamp);
 
+/* OBJECT INSTRUCTIONS */
+
+/**
+ * Sends a body instruction over the given guac_socket connection.
+ *
+ * If an error occurs sending the instruction, a non-zero value is
+ * returned, and guac_error is set appropriately.
+ *
+ * @param socket
+ *     The guac_socket connection to use.
+ *
+ * @param object
+ *     The object to associated with the stream being used.
+ *
+ * @param stream
+ *     The stream to use.
+ *
+ * @param mimetype
+ *     The mimetype of the data being sent.
+ *
+ * @param name
+ *     The name of the stream whose body is being sent, as requested by a "get"
+ *     instruction.
+ *
+ * @return
+ *     Zero on success, non-zero on error.
+ */
+int guac_protocol_send_body(guac_socket* socket, const guac_object* object,
+        const guac_stream* stream, const char* mimetype, const char* name);
+
+/**
+ * Sends a filesystem instruction over the given guac_socket connection.
+ *
+ * If an error occurs sending the instruction, a non-zero value is
+ * returned, and guac_error is set appropriately.
+ *
+ * @param socket
+ *     The guac_socket connection to use.
+ *
+ * @param object
+ *     The object representing the filesystem being exposed.
+ *
+ * @param name
+ *     A name describing the filesystem being exposed.
+ *
+ * @return
+ *     Zero on success, non-zero on error.
+ */
+int guac_protocol_send_filesystem(guac_socket* socket,
+        const guac_object* object, const char* name);
+
+/**
+ * Sends an undefine instruction over the given guac_socket connection.
+ *
+ * If an error occurs sending the instruction, a non-zero value is
+ * returned, and guac_error is set appropriately.
+ *
+ * @param socket
+ *     The guac_socket connection to use.
+ *
+ * @param object
+ *     The object being undefined.
+ *
+ * @return
+ *     Zero on success, non-zero on error.
+ */
+int guac_protocol_send_undefine(guac_socket* socket,
+        const guac_object* object);
+
 /* MEDIA INSTRUCTIONS */
 
 /**
@@ -280,78 +296,14 @@ int guac_protocol_send_sync(guac_socket* socket, guac_timestamp timestamp);
  * returned, and guac_error is set appropriately.
  *
  * @param socket The guac_socket connection to use.
- * @param channel The index of the audio channel the sound should play on.
+ * @param stream The stream to use.
+ * @param channel The index of the audio channel to use.
  * @param mimetype The mimetype of the data being sent.
  * @param duration The duration of the sound being sent, in milliseconds.
- * @param data The audio data to be sent.
- * @param size The number of bytes of audio data to send.
  * @return Zero on success, non-zero on error.
  */
-int guac_protocol_send_audio(guac_socket* socket, int channel,
-        const char* mimetype, double duration, void* data, int size);
-
-/**
- * Begins a audio instruction over the given guac_socket connection. Only the
- * initial non-data part of the instruction and the length of the data part
- * of the instruction are sent. The actual contents of the data must be
- * sent with guac_protocol_send_audio_data(), and the instruction must be
- * completed with guac_protocol_send_audio_end().
- *
- * Note that the size of the audio to be sent MUST be known ahead of time,
- * even though the data of the audio may be sent in chunks.
- *
- * No further instruction data may be sent along the givven guac_socket
- * except via guac_protocol_send_audio_data() until the audio instruction
- * is completed with guac_protocol_send_audio_end().
- *
- * Note that if you send this instruction over a threadsafe socket, you
- * MUST also call guac_protocol_send_audio_end() or the socket will be
- * left in an unsafe state.
- *
- * If an error occurs sending the instruction, a non-zero value is
- * returned, and guac_error is set appropriately.
- *
- * @param socket The guac_socket connection to use.
- * @param channel The index of the audio channel the sound should play on.
- * @param mimetype The mimetype of the data being sent.
- * @param duration The duration of the audio being sent, in milliseconds.
- * @param size The number of bytes of audio data to send.
- * @return Zero on success, non-zero on error.
- */
-int guac_protocol_send_audio_header(guac_socket* socket,
-        int channel, const char* mimetype, double duration, int size);
-
-/**
- * Writes a block of audio data to the currently in-progress audio instruction
- * which was started with guac_protocol_send_audio_header(). Exactly the
- * number of requested bytes are written unless an error occurs. This function
- * may be called multiple times per audio instruction for each chunk of audio
- * data being written, allowing the potentially huge audio instruction to be
- * split across multiple writes.
- *
- * If an error occurs sending the instruction, a non-zero value is
- * returned, and guac_error is set appropriately.
- *
- * @param socket The guac_socket connection to use.
- * @param data The audio data to write.
- * @param count The number of bytes within the given buffer of audio data
- *              that must be written.
- * @return Zero on success, non-zero on error.
- */
-int guac_protocol_send_audio_data(guac_socket* socket, void* data, int count);
-
-/**
- * Completes the audio instruction which was started with
- * guac_protocol_send_audio_header(), and whose data has been written with
- * guac_protocol_send_audio_data().
- *
- * If an error occurs sending the instruction, a non-zero value is
- * returned, and guac_error is set appropriately.
- *
- * @param socket The guac_socket connection to use.
- * @return Zero on success, non-zero on error.
- */
-int guac_protocol_send_audio_end(guac_socket* socket);
+int guac_protocol_send_audio(guac_socket* socket, const guac_stream* stream,
+        int channel, const char* mimetype, double duration);
 
 /**
  * Sends a file instruction over the given guac_socket connection.
@@ -360,13 +312,28 @@ int guac_protocol_send_audio_end(guac_socket* socket);
  * returned, and guac_error is set appropriately.
  *
  * @param socket The guac_socket connection to use.
- * @param index The index of the blob that will contain the contents
- *              of this file.
+ * @param stream The stream to use.
  * @param mimetype The mimetype of the data being sent.
  * @param name A name describing the file being sent.
  * @return Zero on success, non-zero on error.
  */
-int guac_protocol_send_file(guac_socket* socket, int index, const char* mimetype, const char* name);
+int guac_protocol_send_file(guac_socket* socket, const guac_stream* stream,
+        const char* mimetype, const char* name);
+
+/**
+ * Sends a pipe instruction over the given guac_socket connection.
+ *
+ * If an error occurs sending the instruction, a non-zero value is
+ * returned, and guac_error is set appropriately.
+ *
+ * @param socket The guac_socket connection to use.
+ * @param stream The stream to use.
+ * @param mimetype The mimetype of the data being sent.
+ * @param name An arbitrary name uniquely identifying this pipe.
+ * @return Zero on success, non-zero on error.
+ */
+int guac_protocol_send_pipe(guac_socket* socket, const guac_stream* stream,
+        const char* mimetype, const char* name);
 
 /**
  * Writes a block of data to the currently in-progress blob which was already
@@ -376,13 +343,14 @@ int guac_protocol_send_file(guac_socket* socket, int index, const char* mimetype
  * returned, and guac_error is set appropriately.
  *
  * @param socket The guac_socket connection to use.
- * @param index The index of the blob to append data to.
+ * @param stream The stream to use.
  * @param data The file data to write.
  * @param count The number of bytes within the given buffer of file data
  *              that must be written.
  * @return Zero on success, non-zero on error.
  */
-int guac_protocol_send_blob(guac_socket* socket, int index, void* data, int count);
+int guac_protocol_send_blob(guac_socket* socket, const guac_stream* stream,
+        void* data, int count);
 
 /**
  * Sends an end instruction over the given guac_socket connection.
@@ -391,10 +359,10 @@ int guac_protocol_send_blob(guac_socket* socket, int index, void* data, int coun
  * returned, and guac_error is set appropriately.
  *
  * @param socket The guac_socket connection to use.
- * @param index The index of the blob which is now complete.
+ * @param stream The stream to use.
  * @return Zero on success, non-zero on error.
  */
-int guac_protocol_send_end(guac_socket* socket, int index);
+int guac_protocol_send_end(guac_socket* socket, const guac_stream* stream);
 
 /**
  * Sends a video instruction over the given guac_socket connection.
@@ -403,78 +371,14 @@ int guac_protocol_send_end(guac_socket* socket, int index);
  * returned, and guac_error is set appropriately.
  *
  * @param socket The guac_socket connection to use.
+ * @param stream The stream to use.
  * @param layer The destination layer.
  * @param mimetype The mimetype of the data being sent.
  * @param duration The duration of the video being sent, in milliseconds.
- * @param data The video data to be sent.
- * @param size The number of bytes of video data to send.
  * @return Zero on success, non-zero on error.
  */
-int guac_protocol_send_video(guac_socket* socket, const guac_layer* layer,
-        const char* mimetype, double duration, void* data, int size);
-
-/**
- * Begins a video instruction over the given guac_socket connection. Only the
- * initial non-data part of the instruction and the length of the data part
- * of the instruction are sent. The actual contents of the data must be
- * sent with guac_protocol_send_video_data(), and the instruction must be
- * completed with guac_protocol_send_video_end().
- *
- * Note that the size of the video to be sent MUST be known ahead of time,
- * even though the data of the video may be sent in chunks.
- *
- * No further instruction data may be sent along the givven guac_socket
- * except via guac_protocol_send_video_data() until the video instruction
- * is completed with guac_protocol_send_video_end().
- *
- * Note that if you send this instruction over a threadsafe socket, you
- * MUST also call guac_protocol_send_video_end() or the socket will be
- * left in an unsafe state.
- *
- * If an error occurs sending the instruction, a non-zero value is
- * returned, and guac_error is set appropriately.
- *
- * @param socket The guac_socket connection to use.
- * @param layer The destination layer.
- * @param mimetype The mimetype of the data being sent.
- * @param duration The duration of the video being sent, in milliseconds.
- * @param size The number of bytes of video data to send.
- * @return Zero on success, non-zero on error.
- */
-int guac_protocol_send_video_header(guac_socket* socket,
-        const guac_layer* layer, const char* mimetype, double duration, int size);
-
-/**
- * Writes a block of video data to the currently in-progress video instruction
- * which was started with guac_protocol_send_video_header(). Exactly the
- * number of requested bytes are written unless an error occurs. This function
- * may be called multiple times per video instruction for each chunk of video
- * data being written, allowing the potentially huge video instruction to be
- * split across multiple writes.
- *
- * If an error occurs sending the instruction, a non-zero value is
- * returned, and guac_error is set appropriately.
- *
- * @param socket The guac_socket connection to use.
- * @param data The video data to write.
- * @param count The number of bytes within the given buffer of video data
- *              that must be written.
- * @return Zero on success, non-zero on error.
- */
-int guac_protocol_send_video_data(guac_socket* socket, void* data, int count);
-
-/**
- * Completes the video instruction which was started with
- * guac_protocol_send_video_header(), and whose data has been written with
- * guac_protocol_send_video_data().
- *
- * If an error occurs sending the instruction, a non-zero value is
- * returned, and guac_error is set appropriately.
- *
- * @param socket The guac_socket connection to use.
- * @return Zero on success, non-zero on error.
- */
-int guac_protocol_send_video_end(guac_socket* socket);
+int guac_protocol_send_video(guac_socket* socket, const guac_stream* stream,
+        const guac_layer* layer, const char* mimetype, double duration);
 
 /* DRAWING INSTRUCTIONS */
 
@@ -691,22 +595,38 @@ int guac_protocol_send_lstroke(guac_socket* socket,
         const guac_layer* srcl);
 
 /**
- * Sends a png instruction over the given guac_socket connection. The PNG image
- * data given will be automatically base64-encoded for transmission.
+ * Sends an img instruction over the given guac_socket connection.
  *
  * If an error occurs sending the instruction, a non-zero value is
  * returned, and guac_error is set appropriately.
  *
- * @param socket The guac_socket connection to use.
- * @param mode The composite mode to use.
- * @param layer The destination layer.
- * @param x The destination X coordinate.
- * @param y The destination Y coordinate.
- * @param surface A cairo surface containing the image data to send.
- * @return Zero on success, non-zero on error.
+ * @param socket
+ *     The guac_socket connection to use when sending the img instruction.
+ *
+ * @param stream
+ *     The stream over which the image data will be sent.
+ *
+ * @param mode
+ *     The composite mode to use when drawing the image over the destination
+ *     layer.
+ *
+ * @param layer
+ *     The destination layer.
+ *
+ * @param x
+ *     The X coordinate of the upper-left corner of the destination rectangle
+ *     within the destination layer, in pixels.
+ *
+ * @param y
+ *     The Y coordinate of the upper-left corner of the destination rectangle
+ *     within the destination layer, in pixels.
+ *
+ * @return
+ *     Zero if the instruction was successfully sent, non-zero on error.
  */
-int guac_protocol_send_png(guac_socket* socket, guac_composite_mode mode,
-        const guac_layer* layer, int x, int y, cairo_surface_t* surface);
+int guac_protocol_send_img(guac_socket* socket, const guac_stream* stream,
+        guac_composite_mode mode, const guac_layer* layer,
+        const char* mimetype, int x, int y);
 
 /**
  * Sends a pop instruction over the given guac_socket connection.
@@ -912,10 +832,12 @@ int guac_protocol_send_size(guac_socket* socket, const guac_layer* layer,
  * returned, and guac_error is set appropriately.
  *
  * @param socket The guac_socket connection to use.
- * @param data The clipboard data to send.
+ * @param stream The stream to use.
+ * @param mimetype The mimetype of the clipboard data being sent.
  * @return Zero on success, non-zero on error.
  */
-int guac_protocol_send_clipboard(guac_socket* socket, const char* data);
+int guac_protocol_send_clipboard(guac_socket* socket, const guac_stream* stream,
+        const char* mimetype);
 
 /**
  * Sends a name instruction over the given guac_socket connection.
@@ -926,5 +848,14 @@ int guac_protocol_send_clipboard(guac_socket* socket, const char* data);
  */
 int guac_protocol_send_name(guac_socket* socket, const char* name);
 
+/**
+ * Decodes the given base64-encoded string in-place. The base64 string must
+ * be NULL-terminated.
+ *
+ * @param base64 The base64-encoded string to decode.
+ * @return The number of bytes resulting from the decode operation.
+ */
+int guac_protocol_decode_base64(char* base64);
+
 #endif
 
