@@ -1,64 +1,47 @@
+/*
+ * Copyright (C) 2013 Glyptodon LLC
+ *
+ * Permission is hereby granted, free of charge, to any person obtaining a copy
+ * of this software and associated documentation files (the "Software"), to deal
+ * in the Software without restriction, including without limitation the rights
+ * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+ * copies of the Software, and to permit persons to whom the Software is
+ * furnished to do so, subject to the following conditions:
+ *
+ * The above copyright notice and this permission notice shall be included in
+ * all copies or substantial portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+ * THE SOFTWARE.
+ */
 
-/* ***** BEGIN LICENSE BLOCK *****
- * Version: MPL 1.1/GPL 2.0/LGPL 2.1
- *
- * The contents of this file are subject to the Mozilla Public License Version
- * 1.1 (the "License"); you may not use this file except in compliance with
- * the License. You may obtain a copy of the License at
- * http://www.mozilla.org/MPL/
- *
- * Software distributed under the License is distributed on an "AS IS" basis,
- * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
- * for the specific language governing rights and limitations under the
- * License.
- *
- * The Original Code is libguac.
- *
- * The Initial Developer of the Original Code is
- * Michael Jumper.
- * Portions created by the Initial Developer are Copyright (C) 2010
- * the Initial Developer. All Rights Reserved.
- *
- * Contributor(s):
- *
- * Alternatively, the contents of this file may be used under the terms of
- * either the GNU General Public License Version 2 or later (the "GPL"), or
- * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
- * in which case the provisions of the GPL or the LGPL are applicable instead
- * of those above. If you wish to allow use of your version of this file only
- * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the MPL, indicate your
- * decision by deleting the provisions above and replace them with the notice
- * and other provisions required by the GPL or the LGPL. If you do not delete
- * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the MPL, the GPL or the LGPL.
- *
- * ***** END LICENSE BLOCK ***** */
-
-#include <stdlib.h>
-#include <stdio.h>
-#include <stdint.h>
-#include <inttypes.h>
-#include <string.h>
-#include <errno.h>
-
-#ifdef HAVE_PNGSTRUCT_H
-#include <pngstruct.h>
-#endif
-
-#include <png.h>
-
-#include <cairo/cairo.h>
-
-#include <sys/types.h>
+#include "config.h"
 
 #include "error.h"
 #include "layer.h"
+#include "object.h"
 #include "palette.h"
 #include "protocol.h"
 #include "socket.h"
+#include "stream.h"
 #include "unicode.h"
 
+#include <cairo/cairo.h>
+
+#include <inttypes.h>
+#include <setjmp.h>
+#include <stdarg.h>
+#include <stdint.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <sys/types.h>
+
 /* Output formatting functions */
 
 ssize_t __guac_socket_write_length_string(guac_socket* socket, const char* str) {
@@ -70,7 +53,6 @@ ssize_t __guac_socket_write_length_string(guac_socket* socket, const char* str)
 
 }
 
-
 ssize_t __guac_socket_write_length_int(guac_socket* socket, int64_t i) {
 
     char buffer[128];
@@ -79,7 +61,6 @@ ssize_t __guac_socket_write_length_int(guac_socket* socket, int64_t i) {
 
 }
 
-
 ssize_t __guac_socket_write_length_double(guac_socket* socket, double d) {
 
     char buffer[128];
@@ -88,273 +69,28 @@ ssize_t __guac_socket_write_length_double(guac_socket* socket, double d) {
 
 }
 
-
-/* PNG output formatting */
-
-typedef struct __guac_socket_write_png_data {
-
-    guac_socket* socket;
-
-    char* buffer;
-    int buffer_size;
-    int data_size;
-
-} __guac_socket_write_png_data;
-
-cairo_status_t __guac_socket_write_png_cairo(void* closure, const unsigned char* data, unsigned int length) {
-
-    __guac_socket_write_png_data* png_data = (__guac_socket_write_png_data*) closure;
-
-    /* Calculate next buffer size */
-    int next_size = png_data->data_size + length;
-
-    /* If need resizing, double buffer size until big enough */
-    if (next_size > png_data->buffer_size) {
-
-        char* new_buffer;
-
-        do {
-            png_data->buffer_size <<= 1;
-        } while (next_size > png_data->buffer_size);
-
-        /* Resize buffer */
-        new_buffer = realloc(png_data->buffer, png_data->buffer_size);
-        png_data->buffer = new_buffer;
-
-    }
-
-    /* Append data to buffer */
-    memcpy(png_data->buffer + png_data->data_size, data, length);
-    png_data->data_size += length;
-
-    return CAIRO_STATUS_SUCCESS;
-
-}
-
-
-int __guac_socket_write_length_png_cairo(guac_socket* socket, cairo_surface_t* surface) {
-
-    __guac_socket_write_png_data png_data;
-    int base64_length;
-
-    /* Write surface */
-
-    png_data.socket = socket;
-    png_data.buffer_size = 8192;
-    png_data.buffer = malloc(png_data.buffer_size);
-    png_data.data_size = 0;
-
-    if (cairo_surface_write_to_png_stream(surface, __guac_socket_write_png_cairo, &png_data) != CAIRO_STATUS_SUCCESS) {
-        guac_error = GUAC_STATUS_OUTPUT_ERROR;
-        guac_error_message = "Cairo PNG backend failed";
-        return -1;
-    }
-
-    base64_length = (png_data.data_size + 2) / 3 * 4;
-
-    /* Write length and data */
-    if (
-           guac_socket_write_int(socket, base64_length)
-        || guac_socket_write_string(socket, ".")
-        || guac_socket_write_base64(socket, png_data.buffer, png_data.data_size)
-        || guac_socket_flush_base64(socket)) {
-        free(png_data.buffer);
-        return -1;
-    }
-
-    free(png_data.buffer);
-    return 0;
-
-}
-
-void __guac_socket_write_png(png_structp png,
-        png_bytep data, png_size_t length) {
-
-    /* Get png buffer structure */
-    __guac_socket_write_png_data* png_data;
-#ifdef HAVE_PNG_GET_IO_PTR
-    png_data = (__guac_socket_write_png_data*) png_get_io_ptr(png);
-#else
-    png_data = (__guac_socket_write_png_data*) png->io_ptr;
-#endif
-
-    /* Calculate next buffer size */
-    int next_size = png_data->data_size + length;
-
-    /* If need resizing, double buffer size until big enough */
-    if (next_size > png_data->buffer_size) {
-
-        char* new_buffer;
-
-        do {
-            png_data->buffer_size <<= 1;
-        } while (next_size > png_data->buffer_size);
-
-        /* Resize buffer */
-        new_buffer = realloc(png_data->buffer, png_data->buffer_size);
-        png_data->buffer = new_buffer;
-
-    }
-
-    /* Append data to buffer */
-    memcpy(png_data->buffer + png_data->data_size, data, length);
-    png_data->data_size += length;
-
-}
-
-void __guac_socket_flush_png(png_structp png) {
-    /* Dummy function */
-}
-
-int __guac_socket_write_length_png(guac_socket* socket, cairo_surface_t* surface) {
-
-    png_structp png;
-    png_infop png_info;
-    png_byte** png_rows;
-    int bpp;
-
-    int x, y;
-
-    __guac_socket_write_png_data png_data;
-    int base64_length;
-
-    /* Get image surface properties and data */
-    cairo_format_t format = cairo_image_surface_get_format(surface);
-    int width = cairo_image_surface_get_width(surface);
-    int height = cairo_image_surface_get_height(surface);
-    int stride = cairo_image_surface_get_stride(surface);
-    unsigned char* data = cairo_image_surface_get_data(surface);
-
-    /* If not RGB24, use Cairo PNG writer */
-    if (format != CAIRO_FORMAT_RGB24 || data == NULL)
-        return __guac_socket_write_length_png_cairo(socket, surface);
-
-    /* Flush pending operations to surface */
-    cairo_surface_flush(surface);
-
-    /* Attempt to build palette */
-    guac_palette* palette = guac_palette_alloc(surface);
-
-    /* If not possible, resort to Cairo PNG writer */
-    if (palette == NULL)
-        return __guac_socket_write_length_png_cairo(socket, surface);
-
-    /* Calculate BPP from palette size */
-    if      (palette->size <= 2)  bpp = 1;
-    else if (palette->size <= 4)  bpp = 2;
-    else if (palette->size <= 16) bpp = 4;
-    else                          bpp = 8;
-
-    /* Set up PNG writer */
-    png = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
-    if (!png) {
-        guac_error = GUAC_STATUS_OUTPUT_ERROR;
-        guac_error_message = "libpng failed to create write structure";
-        return -1;
-    }
-
-    png_info = png_create_info_struct(png);
-    if (!png_info) {
-        png_destroy_write_struct(&png, NULL);
-        guac_error = GUAC_STATUS_OUTPUT_ERROR;
-        guac_error_message = "libpng failed to create info structure";
-        return -1;
-    }
-
-    /* Set error handler */
-    if (setjmp(png_jmpbuf(png))) {
-        png_destroy_write_struct(&png, &png_info);
-        guac_error = GUAC_STATUS_OUTPUT_ERROR;
-        guac_error_message = "libpng output error";
-        return -1;
-    }
-
-    /* Set up buffer structure */
-    png_data.socket = socket;
-    png_data.buffer_size = 8192;
-    png_data.buffer = malloc(png_data.buffer_size);
-    png_data.data_size = 0;
-
-    /* Set up writer */
-    png_set_write_fn(png, &png_data,
-            __guac_socket_write_png,
-            __guac_socket_flush_png);
-
-    /* Copy data from surface into PNG data */
-    png_rows = (png_byte**) malloc(sizeof(png_byte*) * height);
-    for (y=0; y<height; y++) {
-
-        /* Allocate new PNG row */
-        png_byte* row = (png_byte*) malloc(sizeof(png_byte) * width);
-        png_rows[y] = row;
-
-        /* Copy data from surface into current row */
-        for (x=0; x<width; x++) {
-
-            /* Get pixel color */
-            int color = ((uint32_t*) data)[x] & 0xFFFFFF;
-
-            /* Set index in row */
-            row[x] = guac_palette_find(palette, color);
-
-        }
-
-        /* Advance to next data row */
-        data += stride;
-
-    }
-
-    /* Write image info */
-    png_set_IHDR(
-        png,
-        png_info,
-        width,
-        height,
-        bpp,
-        PNG_COLOR_TYPE_PALETTE,
-        PNG_INTERLACE_NONE,
-        PNG_COMPRESSION_TYPE_DEFAULT,
-        PNG_FILTER_TYPE_DEFAULT
-    );
-
-    /* Write palette */
-    png_set_PLTE(png, png_info, palette->colors, palette->size);
-
-    /* Write image */
-    png_set_rows(png, png_info, png_rows);
-    png_write_png(png, png_info, PNG_TRANSFORM_PACKING, NULL);
-
-    /* Finish write */
-    png_destroy_write_struct(&png, &png_info);
-
-    /* Free palette */
-    guac_palette_free(palette);
-
-    /* Free PNG data */
-    for (y=0; y<height; y++)
-        free(png_rows[y]);
-    free(png_rows);
-
-    base64_length = (png_data.data_size + 2) / 3 * 4;
-
-    /* Write length and data */
-    if (
-           guac_socket_write_int(socket, base64_length)
-        || guac_socket_write_string(socket, ".")
-        || guac_socket_write_base64(socket, png_data.buffer, png_data.data_size)
-        || guac_socket_flush_base64(socket)) {
-        free(png_data.buffer);
-        return -1;
-    }
-
-    free(png_data.buffer);
-    return 0;
-
-}
-
-
 /* Protocol functions */
 
+int guac_protocol_send_ack(guac_socket* socket, guac_stream* stream,
+        const char* error, guac_protocol_status status) {
+
+    int ret_val;
+
+    guac_socket_instruction_begin(socket);
+    ret_val =
+           guac_socket_write_string(socket, "3.ack,")
+        || __guac_socket_write_length_int(socket, stream->index)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_string(socket, error)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_int(socket, status)
+        || guac_socket_write_string(socket, ";");
+
+    guac_socket_instruction_end(socket);
+    return ret_val;
+
+}
+
 static int __guac_protocol_send_args(guac_socket* socket, const char** args) {
 
     int i;
@@ -416,55 +152,30 @@ int guac_protocol_send_arc(guac_socket* socket, const guac_layer* layer,
 
 }
 
-int guac_protocol_send_audio(guac_socket* socket, int channel,
-        const char* mimetype, double duration, void* data, int size) {
+int guac_protocol_send_audio(guac_socket* socket, const guac_stream* stream,
+        int channel, const char* mimetype, double duration) {
 
-    /* By the spec, guac_protocol_send_audio_end() must be called */
-    return
-           (guac_protocol_send_audio_header(socket, channel,
-                mimetype, duration, size)
-        || guac_protocol_send_audio_data(socket, data, size))
-        |  guac_protocol_send_audio_end(socket);
-
-}
-
-int guac_protocol_send_audio_header(guac_socket* socket,
-        int channel, const char* mimetype, double duration, int size) {
-
-    int base64_length = (size + 2) / 3 * 4;
+    int ret_val;
 
     guac_socket_instruction_begin(socket);
-    return 
+    ret_val = 
            guac_socket_write_string(socket, "5.audio,")
+        || __guac_socket_write_length_int(socket, stream->index)
+        || guac_socket_write_string(socket, ",")
         || __guac_socket_write_length_int(socket, channel)
         || guac_socket_write_string(socket, ",")
         || __guac_socket_write_length_string(socket, mimetype)
         || guac_socket_write_string(socket, ",")
         || __guac_socket_write_length_double(socket, duration)
-        || guac_socket_write_string(socket, ",")
-        || guac_socket_write_int(socket, base64_length)
-        || guac_socket_write_string(socket, ".");
-
-}
-
-int guac_protocol_send_audio_data(guac_socket* socket, void* data, int count) {
-
-    return guac_socket_write_base64(socket, data, count);
-
-}
-
-int guac_protocol_send_audio_end(guac_socket* socket) {
-
-    int ret_val =
-           guac_socket_flush_base64(socket)
         || guac_socket_write_string(socket, ";");
-
     guac_socket_instruction_end(socket);
+
     return ret_val;
 
 }
 
-int guac_protocol_send_blob(guac_socket* socket, int index, void* data, int count) {
+int guac_protocol_send_blob(guac_socket* socket, const guac_stream* stream,
+        void* data, int count) {
 
     int base64_length = (count + 2) / 3 * 4;
 
@@ -473,7 +184,7 @@ int guac_protocol_send_blob(guac_socket* socket, int index, void* data, int coun
     guac_socket_instruction_begin(socket);
     ret_val =
            guac_socket_write_string(socket, "4.blob,")
-        || __guac_socket_write_length_int(socket, index)
+        || __guac_socket_write_length_int(socket, stream->index)
         || guac_socket_write_string(socket, ",")
         || guac_socket_write_int(socket, base64_length)
         || guac_socket_write_string(socket, ".")
@@ -486,6 +197,28 @@ int guac_protocol_send_blob(guac_socket* socket, int index, void* data, int coun
 
 }
 
+int guac_protocol_send_body(guac_socket* socket, const guac_object* object,
+        const guac_stream* stream, const char* mimetype, const char* name) {
+
+    int ret_val;
+
+    guac_socket_instruction_begin(socket);
+    ret_val =
+           guac_socket_write_string(socket, "4.body,")
+        || __guac_socket_write_length_int(socket, object->index)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_int(socket, stream->index)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_string(socket, mimetype)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_string(socket, name)
+        || guac_socket_write_string(socket, ";");
+
+    guac_socket_instruction_end(socket);
+    return ret_val;
+
+}
+
 int guac_protocol_send_cfill(guac_socket* socket,
         guac_composite_mode mode, const guac_layer* layer,
         int r, int g, int b, int a) {
@@ -513,7 +246,6 @@ int guac_protocol_send_cfill(guac_socket* socket,
 
 }
 
-
 int guac_protocol_send_close(guac_socket* socket, const guac_layer* layer) {
 
     int ret_val;
@@ -529,7 +261,6 @@ int guac_protocol_send_close(guac_socket* socket, const guac_layer* layer) {
 
 }
 
-
 static int __guac_protocol_send_connect(guac_socket* socket, const char** args) {
 
     int i;
@@ -577,15 +308,17 @@ int guac_protocol_send_clip(guac_socket* socket, const guac_layer* layer) {
 
 }
 
-
-int guac_protocol_send_clipboard(guac_socket* socket, const char* data) {
+int guac_protocol_send_clipboard(guac_socket* socket, const guac_stream* stream,
+        const char* mimetype) {
 
     int ret_val;
 
     guac_socket_instruction_begin(socket);
     ret_val =
            guac_socket_write_string(socket, "9.clipboard,")
-        || __guac_socket_write_length_string(socket, data)
+        || __guac_socket_write_length_int(socket, stream->index)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_string(socket, mimetype)
         || guac_socket_write_string(socket, ";");
 
     guac_socket_instruction_end(socket);
@@ -593,7 +326,6 @@ int guac_protocol_send_clipboard(guac_socket* socket, const char* data) {
 
 }
 
-
 int guac_protocol_send_copy(guac_socket* socket,
         const guac_layer* srcl, int srcx, int srcy, int w, int h,
         guac_composite_mode mode, const guac_layer* dstl, int dstx, int dsty) {
@@ -627,7 +359,6 @@ int guac_protocol_send_copy(guac_socket* socket,
 
 }
 
-
 int guac_protocol_send_cstroke(guac_socket* socket,
         guac_composite_mode mode, const guac_layer* layer,
         guac_line_cap_style cap, guac_line_join_style join, int thickness,
@@ -662,7 +393,6 @@ int guac_protocol_send_cstroke(guac_socket* socket,
 
 }
 
-
 int guac_protocol_send_cursor(guac_socket* socket, int x, int y,
         const guac_layer* srcl, int srcx, int srcy, int w, int h) {
     int ret_val;
@@ -690,7 +420,6 @@ int guac_protocol_send_cursor(guac_socket* socket, int x, int y,
 
 }
 
-
 int guac_protocol_send_curve(guac_socket* socket, const guac_layer* layer,
         int cp1x, int cp1y, int cp2x, int cp2y, int x, int y) {
 
@@ -719,7 +448,6 @@ int guac_protocol_send_curve(guac_socket* socket, const guac_layer* layer,
 
 }
 
-
 int guac_protocol_send_disconnect(guac_socket* socket) {
     int ret_val;
 
@@ -730,7 +458,6 @@ int guac_protocol_send_disconnect(guac_socket* socket) {
 
 }
 
-
 int guac_protocol_send_dispose(guac_socket* socket, const guac_layer* layer) {
 
     int ret_val;
@@ -746,7 +473,6 @@ int guac_protocol_send_dispose(guac_socket* socket, const guac_layer* layer) {
 
 }
 
-
 int guac_protocol_send_distort(guac_socket* socket, const guac_layer* layer,
         double a, double b, double c,
         double d, double e, double f) {
@@ -776,15 +502,14 @@ int guac_protocol_send_distort(guac_socket* socket, const guac_layer* layer,
 
 }
 
-
-int guac_protocol_send_end(guac_socket* socket, int index) {
+int guac_protocol_send_end(guac_socket* socket, const guac_stream* stream) {
 
     int ret_val;
 
     guac_socket_instruction_begin(socket);
     ret_val =
            guac_socket_write_string(socket, "3.end,")
-        || __guac_socket_write_length_int(socket, index)
+        || __guac_socket_write_length_int(socket, stream->index)
         || guac_socket_write_string(socket, ";");
 
     guac_socket_instruction_end(socket);
@@ -792,8 +517,8 @@ int guac_protocol_send_end(guac_socket* socket, int index) {
 
 }
 
-
-int guac_protocol_send_error(guac_socket* socket, const char* error) {
+int guac_protocol_send_error(guac_socket* socket, const char* error,
+        guac_protocol_status status) {
 
     int ret_val;
 
@@ -801,6 +526,8 @@ int guac_protocol_send_error(guac_socket* socket, const char* error) {
     ret_val =
            guac_socket_write_string(socket, "5.error,")
         || __guac_socket_write_length_string(socket, error)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_int(socket, status)
         || guac_socket_write_string(socket, ";");
 
     guac_socket_instruction_end(socket);
@@ -808,15 +535,49 @@ int guac_protocol_send_error(guac_socket* socket, const char* error) {
 
 }
 
+int vguac_protocol_send_log(guac_socket* socket, const char* format,
+        va_list args) {
 
-int guac_protocol_send_file(guac_socket* socket, int index, const char* mimetype, const char* name) {
+    int ret_val;
+
+    /* Copy log message into buffer */
+    char message[4096];
+    vsnprintf(message, sizeof(message), format, args);
+
+    /* Log to instruction */
+    guac_socket_instruction_begin(socket);
+    ret_val =
+           guac_socket_write_string(socket, "3.log,")
+        || __guac_socket_write_length_string(socket, message)
+        || guac_socket_write_string(socket, ";");
+
+    guac_socket_instruction_end(socket);
+    return ret_val;
+
+}
+
+int guac_protocol_send_log(guac_socket* socket, const char* format, ...) {
+
+    int ret_val;
+
+    va_list args;
+    va_start(args, format);
+    ret_val = vguac_protocol_send_log(socket, format, args);
+    va_end(args);
+
+    return ret_val;
+
+}
+
+int guac_protocol_send_file(guac_socket* socket, const guac_stream* stream,
+        const char* mimetype, const char* name) {
 
     int ret_val;
 
     guac_socket_instruction_begin(socket);
     ret_val =
            guac_socket_write_string(socket, "4.file,")
-        || __guac_socket_write_length_int(socket, index)
+        || __guac_socket_write_length_int(socket, stream->index)
         || guac_socket_write_string(socket, ",")
         || __guac_socket_write_length_string(socket, mimetype)
         || guac_socket_write_string(socket, ",")
@@ -828,6 +589,23 @@ int guac_protocol_send_file(guac_socket* socket, int index, const char* mimetype
 
 }
 
+int guac_protocol_send_filesystem(guac_socket* socket,
+        const guac_object* object, const char* name) {
+
+    int ret_val;
+
+    guac_socket_instruction_begin(socket);
+    ret_val =
+           guac_socket_write_string(socket, "10.filesystem,")
+        || __guac_socket_write_length_int(socket, object->index)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_string(socket, name)
+        || guac_socket_write_string(socket, ";");
+
+    guac_socket_instruction_end(socket);
+    return ret_val;
+
+}
 
 int guac_protocol_send_identity(guac_socket* socket, const guac_layer* layer) {
 
@@ -844,7 +622,6 @@ int guac_protocol_send_identity(guac_socket* socket, const guac_layer* layer) {
 
 }
 
-
 int guac_protocol_send_lfill(guac_socket* socket,
         guac_composite_mode mode, const guac_layer* layer,
         const guac_layer* srcl) {
@@ -866,7 +643,6 @@ int guac_protocol_send_lfill(guac_socket* socket,
 
 }
 
-
 int guac_protocol_send_line(guac_socket* socket, const guac_layer* layer,
         int x, int y) {
 
@@ -887,7 +663,6 @@ int guac_protocol_send_line(guac_socket* socket, const guac_layer* layer,
 
 }
 
-
 int guac_protocol_send_lstroke(guac_socket* socket,
         guac_composite_mode mode, const guac_layer* layer,
         guac_line_cap_style cap, guac_line_join_style join, int thickness,
@@ -916,7 +691,6 @@ int guac_protocol_send_lstroke(guac_socket* socket,
 
 }
 
-
 int guac_protocol_send_move(guac_socket* socket, const guac_layer* layer,
         const guac_layer* parent, int x, int y, int z) {
 
@@ -941,7 +715,6 @@ int guac_protocol_send_move(guac_socket* socket, const guac_layer* layer,
 
 }
 
-
 int guac_protocol_send_name(guac_socket* socket, const char* name) {
 
     int ret_val;
@@ -975,23 +748,31 @@ int guac_protocol_send_nest(guac_socket* socket, int index,
 
 }
 
-int guac_protocol_send_png(guac_socket* socket, guac_composite_mode mode,
-        const guac_layer* layer, int x, int y, cairo_surface_t* surface) {
+int guac_protocol_send_nop(guac_socket* socket) {
+
+    int ret_val;
+
+    guac_socket_instruction_begin(socket);
+    ret_val = guac_socket_write_string(socket, "3.nop;");
+    guac_socket_instruction_end(socket);
+
+    return ret_val;
+
+}
+
+int guac_protocol_send_pipe(guac_socket* socket, const guac_stream* stream,
+        const char* mimetype, const char* name) {
 
     int ret_val;
 
     guac_socket_instruction_begin(socket);
     ret_val =
-           guac_socket_write_string(socket, "3.png,")
-        || __guac_socket_write_length_int(socket, mode)
+           guac_socket_write_string(socket, "4.pipe,")
+        || __guac_socket_write_length_int(socket, stream->index)
         || guac_socket_write_string(socket, ",")
-        || __guac_socket_write_length_int(socket, layer->index)
+        || __guac_socket_write_length_string(socket, mimetype)
         || guac_socket_write_string(socket, ",")
-        || __guac_socket_write_length_int(socket, x)
-        || guac_socket_write_string(socket, ",")
-        || __guac_socket_write_length_int(socket, y)
-        || guac_socket_write_string(socket, ",")
-        || __guac_socket_write_length_png(socket, surface)
+        || __guac_socket_write_length_string(socket, name)
         || guac_socket_write_string(socket, ";");
 
     guac_socket_instruction_end(socket);
@@ -999,6 +780,32 @@ int guac_protocol_send_png(guac_socket* socket, guac_composite_mode mode,
 
 }
 
+int guac_protocol_send_img(guac_socket* socket, const guac_stream* stream,
+        guac_composite_mode mode, const guac_layer* layer,
+        const char* mimetype, int x, int y) {
+
+    int ret_val;
+
+    guac_socket_instruction_begin(socket);
+    ret_val =
+           guac_socket_write_string(socket, "3.img,")
+        || __guac_socket_write_length_int(socket, stream->index)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_int(socket, mode)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_int(socket, layer->index)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_string(socket, mimetype)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_int(socket, x)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_int(socket, y)
+        || guac_socket_write_string(socket, ";");
+
+    guac_socket_instruction_end(socket);
+    return ret_val;
+
+}
 
 int guac_protocol_send_pop(guac_socket* socket, const guac_layer* layer) {
 
@@ -1015,7 +822,6 @@ int guac_protocol_send_pop(guac_socket* socket, const guac_layer* layer) {
 
 }
 
-
 int guac_protocol_send_push(guac_socket* socket, const guac_layer* layer) {
 
     int ret_val;
@@ -1031,6 +837,20 @@ int guac_protocol_send_push(guac_socket* socket, const guac_layer* layer) {
 
 }
 
+int guac_protocol_send_ready(guac_socket* socket, const char* id) {
+
+    int ret_val;
+
+    guac_socket_instruction_begin(socket);
+    ret_val =
+           guac_socket_write_string(socket, "5.ready,")
+        || __guac_socket_write_length_string(socket, id)
+        || guac_socket_write_string(socket, ";");
+
+    guac_socket_instruction_end(socket);
+    return ret_val;
+
+}
 
 int guac_protocol_send_rect(guac_socket* socket,
         const guac_layer* layer, int x, int y, int width, int height) {
@@ -1056,7 +876,6 @@ int guac_protocol_send_rect(guac_socket* socket,
 
 }
 
-
 int guac_protocol_send_reset(guac_socket* socket, const guac_layer* layer) {
 
     int ret_val;
@@ -1072,7 +891,6 @@ int guac_protocol_send_reset(guac_socket* socket, const guac_layer* layer) {
 
 }
 
-
 int guac_protocol_send_set(guac_socket* socket, const guac_layer* layer,
         const char* name, const char* value) {
 
@@ -1093,7 +911,6 @@ int guac_protocol_send_set(guac_socket* socket, const guac_layer* layer,
 
 }
 
-
 int guac_protocol_send_select(guac_socket* socket, const char* protocol) {
 
     int ret_val;
@@ -1109,7 +926,6 @@ int guac_protocol_send_select(guac_socket* socket, const char* protocol) {
 
 }
 
-
 int guac_protocol_send_shade(guac_socket* socket, const guac_layer* layer,
         int a) {
 
@@ -1128,7 +944,6 @@ int guac_protocol_send_shade(guac_socket* socket, const guac_layer* layer,
 
 }
 
-
 int guac_protocol_send_size(guac_socket* socket, const guac_layer* layer,
         int w, int h) {
 
@@ -1149,7 +964,6 @@ int guac_protocol_send_size(guac_socket* socket, const guac_layer* layer,
 
 }
 
-
 int guac_protocol_send_start(guac_socket* socket, const guac_layer* layer,
         int x, int y) {
 
@@ -1170,7 +984,6 @@ int guac_protocol_send_start(guac_socket* socket, const guac_layer* layer,
 
 }
 
-
 int guac_protocol_send_sync(guac_socket* socket, guac_timestamp timestamp) {
 
     int ret_val;
@@ -1186,7 +999,6 @@ int guac_protocol_send_sync(guac_socket* socket, guac_timestamp timestamp) {
 
 }
 
-
 int guac_protocol_send_transfer(guac_socket* socket,
         const guac_layer* srcl, int srcx, int srcy, int w, int h,
         guac_transfer_function fn, const guac_layer* dstl, int dstx, int dsty) {
@@ -1220,7 +1032,6 @@ int guac_protocol_send_transfer(guac_socket* socket,
 
 }
 
-
 int guac_protocol_send_transform(guac_socket* socket, const guac_layer* layer,
         double a, double b, double c,
         double d, double e, double f) {
@@ -1250,47 +1061,15 @@ int guac_protocol_send_transform(guac_socket* socket, const guac_layer* layer,
 
 }
 
-int guac_protocol_send_video(guac_socket* socket, const guac_layer* layer,
-        const char* mimetype, double duration, void* data, int size) {
+int guac_protocol_send_undefine(guac_socket* socket,
+        const guac_object* object) {
 
-    /* By the spec, guac_protocol_send_video_end() must be called */
-    return
-           (guac_protocol_send_video_header(socket, layer,
-                mimetype, duration, size)
-        || guac_protocol_send_video_data(socket, data, size))
-        |  guac_protocol_send_video_end(socket);
-
-}
-
-int guac_protocol_send_video_header(guac_socket* socket,
-        const guac_layer* layer, const char* mimetype, double duration, int size) {
-
-    int base64_length = (size + 2) / 3 * 4;
+    int ret_val;
 
     guac_socket_instruction_begin(socket);
-    return
-           guac_socket_write_string(socket, "5.video,")
-        || __guac_socket_write_length_int(socket, layer->index)
-        || guac_socket_write_string(socket, ",")
-        || __guac_socket_write_length_string(socket, mimetype)
-        || guac_socket_write_string(socket, ",")
-        || __guac_socket_write_length_double(socket, duration)
-        || guac_socket_write_string(socket, ",")
-        || guac_socket_write_int(socket, base64_length)
-        || guac_socket_write_string(socket, ".");
-
-}
-
-int guac_protocol_send_video_data(guac_socket* socket, void* data, int count) {
-
-    return guac_socket_write_base64(socket, data, count);
-
-}
-
-int guac_protocol_send_video_end(guac_socket* socket) {
-
-    int ret_val =
-           guac_socket_flush_base64(socket)
+    ret_val =
+           guac_socket_write_string(socket, "8.undefine,")
+        || __guac_socket_write_length_int(socket, object->index)
         || guac_socket_write_string(socket, ";");
 
     guac_socket_instruction_end(socket);
@@ -1298,3 +1077,84 @@ int guac_protocol_send_video_end(guac_socket* socket) {
 
 }
 
+int guac_protocol_send_video(guac_socket* socket, const guac_stream* stream,
+        const guac_layer* layer, const char* mimetype, double duration) {
+
+    int ret_val;
+
+    guac_socket_instruction_begin(socket);
+    ret_val = 
+           guac_socket_write_string(socket, "5.video,")
+        || __guac_socket_write_length_int(socket, stream->index)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_int(socket, layer->index)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_string(socket, mimetype)
+        || guac_socket_write_string(socket, ",")
+        || __guac_socket_write_length_double(socket, duration)
+        || guac_socket_write_string(socket, ";");
+    guac_socket_instruction_end(socket);
+
+    return ret_val;
+
+}
+
+/**
+ * Returns the value of a single base64 character.
+ */
+static int __guac_base64_value(char c) {
+
+    if (c >= 'A' && c <= 'Z')
+        return c - 'A';
+
+    if (c >= 'a' && c <= 'z')
+        return c - 'a' + 26;
+
+    if (c >= '0' && c <= '9')
+        return c - '0' + 52;
+
+    if (c == '+')
+        return 62;
+
+    if (c == '/')
+        return 63;
+
+    return 0;
+
+}
+
+int guac_protocol_decode_base64(char* base64) {
+
+    char* input = base64;
+    char* output = base64;
+
+    int length = 0;
+    int bits_read = 0;
+    int value = 0;
+    char current;
+
+    /* For all characters in string */
+    while ((current = *(input++)) != 0) {
+
+        /* If we've reached padding, then we're done */
+        if (current == '=')
+            break;
+
+        /* Otherwise, shift on the latest 6 bits */
+        value = (value << 6) | __guac_base64_value(current);
+        bits_read += 6;
+
+        /* If we have at least one byte, write out the latest whole byte */
+        if (bits_read >= 8) {
+            *(output++) = (value >> (bits_read % 8)) & 0xFF;
+            bits_read -= 8;
+            length++;
+        }
+
+    }
+
+    /* Return number of bytes written */
+    return length;
+
+}
+
